"""
.. _neomodel_generate_diagram:

``neomodel_generate_diagram``
-----------------------------

::

    usage: _neomodel_generate_diagram [-h] [--file-type <arrows|puml>] [--write-to-dir <someapp/diagrams> ...]
    
    Connects to a Neo4j database and inspects existing nodes and relationships.
    Infers the schema of the database and generates Python class definitions.

    If a connection URL is not specified, the tool will look up the environment 
    variable NEO4J_BOLT_URL. If that environment variable is not set, the tool
    will attempt to connect to the default URL bolt://neo4j:neo4j@localhost:7687

    If a file is specified, the tool will write the class definitions to that file.
    If no file is specified, the tool will print the class definitions to stdout.

    Note : this script only has a synchronous mode.
    
    options:
        -h, --help            show this help message and exit
        -T, --file-type <arrows|puml>
                            File type to produce. Accepts PlantUML (puml) or Arrows.app (arrows). Default is Arrows.
        -D, --write-to-dir someapp/diagrams
                            Directory where to write output file. Default is current directory.
"""

import argparse
import json
import math
import os
import textwrap

from neomodel import (
    ArrayProperty,
    AsyncRelationshipFrom,
    AsyncRelationshipTo,
    AsyncStructuredNode,
    BooleanProperty,
    DateProperty,
    DateTimeFormatProperty,
    DateTimeNeo4jFormatProperty,
    DateTimeProperty,
    FloatProperty,
    IntegerProperty,
    RelationshipFrom,
    RelationshipTo,
    StringProperty,
    StructuredNode,
    UniqueIdProperty,
)
from neomodel.contrib import AsyncSemiStructuredNode, SemiStructuredNode
from neomodel.contrib.spatial_properties import PointProperty
from neomodel.scripts.utils import load_python_module_or_file, recursive_list_classes


def generate_plantuml(classes):
    filename = "model_diagram.puml"
    diagram = "@startuml\n"

    dot_output = "digraph G {\n"
    dot_output += "  node [shape=record];\n"
    for cls in classes:
        # Node label construction for properties
        label = f"{cls.__name__}|{{"
        properties = [
            f"{prop}: {parse_property_key(cls.defined_properties(aliases=False, rels=False)[prop])}"
            for prop in cls.defined_properties(aliases=False, rels=False)
        ]
        label += r" \l ".join(properties)
        label += "}}"

        # Node definition
        dot_output += f'  {cls.__name__} [label="{label}"];\n'

        # Relationships
        for rel_name, rel in cls.defined_properties(
            aliases=False, properties=False
        ).items():
            target_cls = rel._raw_class
            edge_label = f"{rel_name}: {rel.__class__.__name__}"
            if isinstance(rel, RelationshipTo) or isinstance(rel, AsyncRelationshipTo):
                dot_output += (
                    f'  {cls.__name__} -> {target_cls} [label="{edge_label}"];\n'
                )
            elif isinstance(rel, RelationshipFrom) or isinstance(
                rel, AsyncRelationshipFrom
            ):
                dot_output += (
                    f'  {target_cls} -> {cls.__name__} [label="{edge_label}"];\n'
                )

    dot_output += "}"
    diagram += dot_output
    diagram += "@enduml"
    return filename, diagram


def transform_property_type(prop_definition):
    if isinstance(prop_definition, StringProperty):
        return "str"
    elif isinstance(prop_definition, UniqueIdProperty):
        return "id"
    elif isinstance(prop_definition, BooleanProperty):
        return "bool"
    elif isinstance(prop_definition, DateProperty):
        return "date"
    elif (
        isinstance(prop_definition, DateTimeProperty)
        or isinstance(prop_definition, DateTimeFormatProperty)
        or isinstance(prop_definition, DateTimeNeo4jFormatProperty)
    ):
        return "datetime"
    elif isinstance(prop_definition, IntegerProperty):
        return "int"
    elif isinstance(prop_definition, FloatProperty):
        return "float"
    elif isinstance(prop_definition, ArrayProperty):
        return f"list[{transform_property_type(prop_definition.base_property)}]"
    elif isinstance(prop_definition, PointProperty):
        return "point"


def parse_property_key(prop_definition):
    output = transform_property_type(prop_definition)

    if (
        prop_definition.required
        or prop_definition.index
        or prop_definition.unique_index
    ):
        output += " - "
        suffixes = []
        if prop_definition.required:
            suffixes.append("required")
        elif prop_definition.unique_index:
            suffixes.append("unique")
        if prop_definition.index:
            suffixes.append("index")
        output += ", ".join(suffixes)
    return output


def generate_arrows_json(classes):
    filename = "model_diagram.json"
    nodes = []
    edges = []
    positions = {"x": 0, "y": 0}
    radius_increment = 400  # Horizontal space between nodes

    for idx, cls in enumerate(classes):
        node_id = f"n{idx}"
        # Set positions such that related nodes are close on the y-axis
        position = {"x": positions["x"], "y": positions["y"]}
        if idx != 0 and idx % 6 == 0:
            radius_increment += radius_increment
            positions["x"] += radius_increment
            positions["y"] = 0
        else:
            angle = (idx % 6) * (2 * math.pi / 6) + (math.pi / 6)
            if idx % 12 > 6:
                angle += math.pi / 6
            positions["x"] = radius_increment * math.cos(angle)
            positions["y"] = radius_increment * math.sin(angle)

        nodes.append(
            {
                "id": node_id,
                "position": position,
                "caption": "",
                "style": {},
                "labels": [cls.__name__],
                "properties": {
                    prop: parse_property_key(
                        cls.defined_properties(aliases=False, rels=False)[prop]
                    )
                    for prop in cls.defined_properties(aliases=False, rels=False)
                },
            }
        )

        # Prepare relationships
        for _, rel in cls.defined_properties(aliases=False, properties=False).items():
            target_cls = [
                _class for _class in classes if _class.__name__ == rel._raw_class
            ][0]
            target_idx = classes.index(target_cls)
            target_id = f"n{target_idx}"
            # Create edges
            edges.append(
                {
                    "id": f"e{len(edges)}",
                    "type": rel.definition["relation_type"],
                    "style": {},
                    "properties": {},
                    "fromId": (
                        node_id
                        if (
                            isinstance(rel, RelationshipTo)
                            or isinstance(rel, AsyncRelationshipTo)
                        )
                        else target_id
                    ),
                    "toId": (
                        target_id
                        if (
                            isinstance(rel, RelationshipTo)
                            or isinstance(rel, AsyncRelationshipTo)
                        )
                        else node_id
                    ),
                }
            )

    return filename, json.dumps(
        {
            "style": {
                "node-color": "#ffffff",
                "border-color": "#000000",
                "caption-color": "#000000",
                "arrow-color": "#000000",
                "label-background-color": "#ffffff",
                "directionality": "directed",
                "arrow-width": 5,
            },
            "nodes": nodes,
            "relationships": edges,
        },
        indent=4,
    )


def main():
    # Parse command line arguments
    parser = argparse.ArgumentParser(
        description=textwrap.dedent(
            """
            Connects to a Neo4j database and inspects existing nodes and relationships.
            Infers the schema of the database and generates Python class definitions.

            If a connection URL is not specified, the tool will look up the environment 
            variable NEO4J_BOLT_URL. If that environment variable is not set, the tool
            will attempt to connect to the default URL bolt://neo4j:neo4j@localhost:7687

            If a file is specified, the tool will write the class definitions to that file.
            If no file is specified, the tool will print the class definitions to stdout.
            """
        ),
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument(
        "apps",
        metavar="<someapp/models/app.py>",
        type=str,
        nargs="+",
        help="python modules or files with neomodel schema declarations.",
    )

    parser.add_argument(
        "-T",
        "--file-type",
        metavar="<arrows|puml>",
        type=str,
        default="arrows",
        help="File type to produce. Accepts : [arrows, puml]. Default is arrows.",
    )

    parser.add_argument(
        "-D",
        "--write-to-dir",
        metavar="someapp/diagrams",
        type=str,
        default=".",
        help="Directory where to write output file. Default is current directory.",
    )

    args = parser.parse_args()

    for app in args.apps:
        load_python_module_or_file(app)

    classes = recursive_list_classes(StructuredNode, exclude_list=[SemiStructuredNode])
    classes += recursive_list_classes(
        AsyncStructuredNode, exclude_list=[AsyncSemiStructuredNode]
    )

    filename = ""
    output = ""
    if args.file_type == "puml":
        filename, output = generate_plantuml(classes)
    elif args.file_type == "arrows":
        filename, output = generate_arrows_json(classes)
    else:
        raise ValueError(f"Unsupported file type : {args.file_type}")

    # Save to a file
    with open(os.path.join(args.write_to_dir, filename), "w", encoding="utf-8") as file:
        file.write(output)
        print("Successfully wrote diagram to file : ", file.name)


if __name__ == "__main__":
    main()



from neomodel.contrib.spatial_properties import NeomodelPoint
import pytest

def test_failed_initialisation():
    """
    Tests that a NeomodelPoint cannot be instantiated in an erroneous state
    """

    # Must be instantiated with a meaningful CRS
    with pytest.raises(ValueError):
        p1 = NeomodelPoint(crs="blah")

    # Must be instantiated with the right dimensionality of coordinates (either 2 or 3)
    with pytest.raises(ValueError):
        p1 = NeomodelPoint([0,0,0,0])

    with pytest.raises(ValueError):
        p1 = NeomodelPoint([0,])

    # Copy constructor only on the right type of object
    with pytest.raises(TypeError):
        p1 = NeomodelPoint("Definitely not a Neomodel point")

    # Cannot be instantiated in a cartesian or geographical CRS at the same time
    with pytest.raises(ValueError):
        p1 = NeomodelPoint(x=0, y=0, z=0, longitude=0, latitude=0, height=0)

    # Cannot be instantiated without actually pointing somehwere in space
    with pytest.raises(ValueError):
        p1 = NeomodelPoint()

    # CRS and supplied arguments should match (3 for a 3d crs, 2 for a 2d crs)
    with pytest.raises(ValueError):
        p1 = NeomodelPoint(x=0, y=0, z=0, crs="cartesian")

    with pytest.raises(ValueError):
        p1 = NeomodelPoint(x=0, y=0, crs="cartesian-3d")

def test_succesful_initialisation():
    """
    Expected initialisation and copy constructor
    """

    p1 = NeomodelPoint(x=0, y=0)
    p2 = NeomodelPoint(x=0, y=0, z=0)
    p3 = NeomodelPoint(longitude=0, latitude=0)
    p4 = NeomodelPoint(longitude=0, latitude=0, height=0)
    p5 = NeomodelPoint([0,0])
    p6 = NeomodelPoint([0,0,0])
    p7 = NeomodelPoint(p6)

    assert p1.crs=="cartesian" and p1.x==0 and p1.y==0
    assert p2.crs=="cartesian-3d" and p2.x==0 and p2.y==0 and p2.z==0

    assert p5.crs=="cartesian" and p5.x==0 and p5.y==0
    assert p6.crs=="cartesian-3d" and p6.x==0 and p6.y==0 and p6.z==0

    assert p3.crs=="wgs-84" and p3.longitude==0 and p3.latitude==0
    assert p4.crs=="wgs-84-3d" and p4.longitude==0 and p4.latitude==0 and p4.height==0

    assert p7.crs=="cartesian-3d" and p7.x==0 and p7.y==0 and p7.z==0

def test_property_access():
    """
    Points initialised as 2d cannot offer access to 3d coordinates
    """

    p1 = NeomodelPoint(x=0,y=0)
    p2 = NeomodelPoint(longitude=0, latitude=0, height=0)

    with pytest.raises(TypeError):
        assert p1.longitude == 0

    with pytest.raises(TypeError):
        assert p1.latitude == 0

    with pytest.raises(TypeError):
        assert p1.height == 0

    with pytest.raises(TypeError):
        assert p2.x == 0

    with pytest.raises(TypeError):
        assert p2.y == 0

    with pytest.raises(TypeError):
        assert p2.z == 0

def test_equality_success():
    """
    Points with identical coordinates and CRS are equal in value
    """
    p1 = NeomodelPoint(x=0, y=0, z=0)
    p2 = NeomodelPoint(x=0, y=0, z=0)

    assert p1 == p2

def test_equality_fails():
    """
    Points are comparable only with points
    """
    p1 = NeomodelPoint(x=0, y=0, z=0)

    with pytest.raises(ValueError):
        assert p1 == 4


def test_equality_successful():
    """
    Points with identical coordinates and CRS are equal in value
    """
    p1 = NeomodelPoint(x=0, y=0, z=0)
    p2 = NeomodelPoint(x=0, y=0, z=0)
    p3 = NeomodelPoint(longitude=0, latitude=0, height=0)
    p4 = NeomodelPoint(longitude=0, latitude=0, height=0)

    assert p1 == p2
    assert p3 == p4
    assert p1!=p4
    assert p3!=p2





from test._async_compat import mark_async_test
from test.async_.test_relationships import Country

from neomodel import AsyncRelationshipTo, AsyncStructuredNode, StringProperty


class Cat(AsyncStructuredNode):
    name = StringProperty()
    # Relationship is defined using a relative class path
    is_from = AsyncRelationshipTo(".test_relationships.Country", "IS_FROM")


@mark_async_test
async def test_relative_relationship():
    a = await Cat(name="snufkin").save()
    assert a

    c = await Country(code="MG").save()
    assert c

    # connecting an instance of the class defined above
    # the next statement will fail if there's a type mismatch
    await a.is_from.connect(c)
    assert await a.is_from.is_connected(c)



import neo4j

from neomodel._version import __version__

# Use this to connect with automatically created driver
# The following options are the default ones that will be used as driver config
DATABASE_URL = "bolt://neo4j:foobarbaz@localhost:7687"
FORCE_TIMEZONE = False

CONNECTION_ACQUISITION_TIMEOUT = 60.0
CONNECTION_TIMEOUT = 30.0
ENCRYPTED = False
KEEP_ALIVE = True
MAX_CONNECTION_LIFETIME = 3600
MAX_CONNECTION_POOL_SIZE = 100
MAX_TRANSACTION_RETRY_TIME = 30.0
RESOLVER = None
TRUSTED_CERTIFICATES = neo4j.TrustSystemCAs()
USER_AGENT = f"neomodel/v{__version__}"

# Use this to connect with your self-managed driver instead
# DRIVER = neo4j.GraphDatabase().driver(
#     "bolt://localhost:7687", auth=("neo4j", "foobarbaz")
# )
DRIVER = None
# Use this to connect to a specific database when using the self-managed driver
DATABASE_NAME = None

# Use this to enable soft cardinality check
SOFT_INVERSE_CARDINALITY_CHECK = True



from functools import wraps
from typing import Any, Callable


def _exec_hook(hook_name: str, self: Any) -> None:
    if hasattr(self, hook_name):
        getattr(self, hook_name)()


def hooks(fn: Callable) -> Callable:
    @wraps(fn)
    def hooked(self: Any) -> Callable:
        fn_name = getattr(fn, "func_name", fn.__name__)
        _exec_hook("pre_" + fn_name, self)
        val = fn(self)
        _exec_hook("post_" + fn_name, self)
        return val

    return hooked



from typing import TYPE_CHECKING, Any, Optional

from neomodel.exceptions import AttemptedCardinalityViolation, CardinalityViolation
from neomodel.sync_.relationship_manager import (  # pylint:disable=unused-import
    RelationshipManager,
    ZeroOrMore,
)

if TYPE_CHECKING:
    from neomodel import StructuredNode, StructuredRel


class ZeroOrOne(RelationshipManager):
    """A relationship to zero or one node."""

    description = "zero or one relationship"

    def _check_cardinality(
        self, node: "StructuredNode", soft_check: bool = False
    ) -> None:
        if self.__len__():
            if soft_check:
                print(
                    f"Cardinality violation detected : Node already has one relationship of type {self.definition['relation_type']}, should not connect more. Soft check is enabled so the relationship will be created. Note that strict check will be enabled by default in version 6.0"
                )
            else:
                raise AttemptedCardinalityViolation(
                    f"Node already has one relationship of type {self.definition['relation_type']}. Use reconnect() to replace the existing relationship."
                )

    def single(self) -> Optional["StructuredNode"]:
        """
        Return the associated node.

        :return: node
        """
        nodes = super().all()
        if len(nodes) == 1:
            return nodes[0]
        if len(nodes) > 1:
            raise CardinalityViolation(self, len(nodes))
        return None

    def all(self) -> list["StructuredNode"]:
        node = self.single()
        return [node] if node else []

    def connect(
        self, node: "StructuredNode", properties: Optional[dict[str, Any]] = None
    ) -> "StructuredRel":
        """
        Connect to a node.

        :param node:
        :type: StructuredNode
        :param properties: relationship properties
        :type: dict
        :return: True / rel instance
        """
        return super().connect(node, properties)


class OneOrMore(RelationshipManager):
    """A relationship to zero or more nodes."""

    description = "one or more relationships"

    def single(self) -> "StructuredNode":
        """
        Fetch one of the related nodes

        :return: Node
        """
        nodes = super().all()
        if nodes:
            return nodes[0]
        raise CardinalityViolation(self, "none")

    def all(self) -> list["StructuredNode"]:
        """
        Returns all related nodes.

        :return: [node1, node2...]
        """
        nodes = super().all()
        if nodes:
            return nodes
        raise CardinalityViolation(self, "none")

    def disconnect(self, node: "StructuredNode") -> None:
        """
        Disconnect node
        :param node:
        :return:
        """
        if super().__len__() < 2:
            raise AttemptedCardinalityViolation("One or more expected")
        return super().disconnect(node)


class One(RelationshipManager):
    """
    A relationship to a single node
    """

    description = "one relationship"

    def _check_cardinality(
        self, node: "StructuredNode", soft_check: bool = False
    ) -> None:
        if self.__len__():
            if soft_check:
                print(
                    f"Cardinality violation detected : Node already has one relationship of type {self.definition['relation_type']}, should not connect more. Soft check is enabled so the relationship will be created. Note that strict check will be enabled by default in version 6.0"
                )
            else:
                raise AttemptedCardinalityViolation(
                    f"Node already has one relationship of type {self.definition['relation_type']}. Use reconnect() to replace the existing relationship."
                )

    def single(self) -> "StructuredNode":
        """
        Return the associated node.

        :return: node
        """
        nodes = super().all()
        if nodes:
            if len(nodes) == 1:
                return nodes[0]
            raise CardinalityViolation(self, len(nodes))
        raise CardinalityViolation(self, "none")

    def all(self) -> list["StructuredNode"]:
        """
        Return single node in an array

        :return: [node]
        """
        return [self.single()]

    def disconnect(self, node: "StructuredNode") -> None:
        raise AttemptedCardinalityViolation(
            "Cardinality one, cannot disconnect use reconnect."
        )

    def disconnect_all(self) -> None:
        raise AttemptedCardinalityViolation(
            "Cardinality one, cannot disconnect_all use reconnect."
        )

    def connect(
        self, node: "StructuredNode", properties: Optional[dict[str, Any]] = None
    ) -> "StructuredRel":
        """
        Connect a node

        :param node:
        :param properties: relationship properties
        :return: True / rel instance
        """
        if not hasattr(self.source, "element_id") or self.source.element_id is None:
            raise ValueError("Node has not been saved cannot connect!")
        return super().connect(node, properties)



from test._async_compat import mark_sync_test

from pytest import raises

from neomodel import (
    IntegerProperty,
    One,
    Q,
    Relationship,
    RelationshipFrom,
    RelationshipTo,
    StringProperty,
    StructuredNode,
    StructuredRel,
    db,
)


class PersonWithRels(StructuredNode):
    name = StringProperty(unique_index=True)
    age = IntegerProperty(index=True)
    is_from = RelationshipTo("Country", "IS_FROM")
    knows = Relationship("PersonWithRels", "KNOWS")

    @property
    def special_name(self):
        return self.name

    def special_power(self):
        return "I have no powers"


class Country(StructuredNode):
    code = StringProperty(unique_index=True)
    inhabitant = RelationshipFrom(PersonWithRels, "IS_FROM")
    president = RelationshipTo(PersonWithRels, "PRESIDENT", cardinality=One)


class SuperHero(PersonWithRels):
    power = StringProperty(index=True)

    def special_power(self):
        return "I have powers"


@mark_sync_test
def test_actions_on_deleted_node():
    u = PersonWithRels(name="Jim2", age=3).save()
    u.delete()
    with raises(ValueError):
        u.is_from.connect(None)

    with raises(ValueError):
        u.is_from.get()

    with raises(ValueError):
        u.save()


@mark_sync_test
def test_bidirectional_relationships():
    u = PersonWithRels(name="Jim", age=3).save()
    assert u

    de = Country(code="DE").save()
    assert de

    assert not u.is_from.all()

    assert u.is_from.__class__.__name__ == "ZeroOrMore"
    u.is_from.connect(de)

    assert len(u.is_from.all()) == 1

    assert u.is_from.is_connected(de)

    b = (u.is_from.all())[0]
    assert b.__class__.__name__ == "Country"
    assert b.code == "DE"

    s = (b.inhabitant.all())[0]
    assert s.name == "Jim"

    u.is_from.disconnect(b)
    assert not u.is_from.is_connected(b)


@mark_sync_test
def test_either_direction_connect():
    rey = PersonWithRels(name="Rey", age=3).save()
    sakis = PersonWithRels(name="Sakis", age=3).save()

    rey.knows.connect(sakis)
    assert rey.knows.is_connected(sakis)
    assert sakis.knows.is_connected(rey)
    sakis.knows.connect(rey)

    result, _ = sakis.cypher(
        f"""MATCH (us), (them)
            WHERE {db.get_id_method()}(us)=$self and {db.get_id_method()}(them)=$them
            MATCH (us)-[r:KNOWS]-(them) RETURN COUNT(r)""",
        {"them": db.parse_element_id(rey.element_id)},
    )
    assert int(result[0][0]) == 1

    rel = rey.knows.relationship(sakis)
    assert isinstance(rel, StructuredRel)

    rels = rey.knows.all_relationships(sakis)
    assert isinstance(rels[0], StructuredRel)


@mark_sync_test
def test_search_and_filter_and_exclude():
    fred = PersonWithRels(name="Fred", age=13).save()
    zz = Country(code="ZZ").save()
    zx = Country(code="ZX").save()
    zt = Country(code="ZY").save()
    fred.is_from.connect(zz)
    fred.is_from.connect(zx)
    fred.is_from.connect(zt)
    result = fred.is_from.filter(code="ZX")
    assert result[0].code == "ZX"

    result = fred.is_from.filter(code="ZY")
    assert result[0].code == "ZY"

    result = fred.is_from.exclude(code="ZZ").exclude(code="ZY")
    assert result[0].code == "ZX" and len(result) == 1

    result = fred.is_from.exclude(Q(code__contains="Y"))
    assert len(result) == 2

    result = fred.is_from.filter(Q(code__contains="Z"))
    assert len(result) == 3


@mark_sync_test
def test_custom_methods():
    u = PersonWithRels(name="Joe90", age=13).save()
    assert u.special_power() == "I have no powers"
    u = SuperHero(name="Joe91", age=13, power="xxx").save()
    assert u.special_power() == "I have powers"
    assert u.special_name == "Joe91"


@mark_sync_test
def test_valid_reconnection():
    p = PersonWithRels(name="ElPresidente", age=93).save()
    assert p

    pp = PersonWithRels(name="TheAdversary", age=33).save()
    assert pp

    c = Country(code="CU").save()
    assert c

    c.president.connect(p)
    assert c.president.is_connected(p)

    # the coup d'etat
    c.president.reconnect(p, pp)
    assert c.president.is_connected(pp)

    # reelection time
    c.president.reconnect(pp, pp)
    assert c.president.is_connected(pp)


@mark_sync_test
def test_valid_replace():
    brady = PersonWithRels(name="Tom Brady", age=40).save()
    assert brady

    gronk = PersonWithRels(name="Rob Gronkowski", age=28).save()
    assert gronk

    colbert = PersonWithRels(name="Stephen Colbert", age=53).save()
    assert colbert

    hanks = PersonWithRels(name="Tom Hanks", age=61).save()
    assert hanks

    brady.knows.connect(gronk)
    brady.knows.connect(colbert)
    assert len(brady.knows.all()) == 2
    assert brady.knows.is_connected(gronk)
    assert brady.knows.is_connected(colbert)

    brady.knows.replace(hanks)
    assert len(brady.knows.all()) == 1
    assert brady.knows.is_connected(hanks)
    assert not brady.knows.is_connected(gronk)
    assert not brady.knows.is_connected(colbert)


@mark_sync_test
def test_props_relationship():
    u = PersonWithRels(name="Mar", age=20).save()
    assert u

    c = Country(code="AT").save()
    assert c

    c2 = Country(code="LA").save()
    assert c2

    with raises(NotImplementedError):
        c.inhabitant.connect(u, properties={"city": "Thessaloniki"})



from test._async_compat import mark_async_test

from neo4j.exceptions import ClientError

from neomodel import AsyncStructuredNode, StringProperty, adb


class ConstraintAndIndex(AsyncStructuredNode):
    name = StringProperty(unique_index=True)
    last_name = StringProperty(index=True)


@mark_async_test
async def test_drop_labels():
    await adb.install_labels(ConstraintAndIndex)
    constraints_before = await adb.list_constraints()
    indexes_before = await adb.list_indexes(exclude_token_lookup=True)

    assert len(constraints_before) > 0
    assert len(indexes_before) > 0

    await adb.remove_all_labels()

    constraints = await adb.list_constraints()
    indexes = await adb.list_indexes(exclude_token_lookup=True)

    assert len(constraints) == 0
    assert len(indexes) == 0

    # Recreating all old constraints and indexes
    for constraint in constraints_before:
        constraint_type_clause = "UNIQUE"
        if constraint["type"] == "NODE_PROPERTY_EXISTENCE":
            constraint_type_clause = "NOT NULL"
        elif constraint["type"] == "NODE_KEY":
            constraint_type_clause = "NODE KEY"

        await adb.cypher_query(
            f'CREATE CONSTRAINT {constraint["name"]} FOR (n:{constraint["labelsOrTypes"][0]}) REQUIRE n.{constraint["properties"][0]} IS {constraint_type_clause}'
        )
    for index in indexes_before:
        try:
            await adb.cypher_query(
                f'CREATE INDEX {index["name"]} FOR (n:{index["labelsOrTypes"][0]}) ON (n.{index["properties"][0]})'
            )
        except ClientError:
            pass



from test._async_compat import mark_sync_test

import pytest

from neomodel import (
    IntegerProperty,
    RelationshipTo,
    StringProperty,
    StructuredNode,
    StructuredRel,
    db,
)


class Album(StructuredNode):
    name = StringProperty()


class Released(StructuredRel):
    year = IntegerProperty()


class Band(StructuredNode):
    name = StringProperty()
    released = RelationshipTo(Album, relation_type="RELEASED", model=Released)


@mark_sync_test
def test_read_elements_id():
    the_hives = Band(name="The Hives").save()
    lex_hives = Album(name="Lex Hives").save()
    released_rel = the_hives.released.connect(lex_hives)

    # Validate element_id properties
    assert lex_hives.element_id == (the_hives.released.single()).element_id
    assert released_rel._start_node_element_id == the_hives.element_id
    assert released_rel._end_node_element_id == lex_hives.element_id

    # Validate id properties
    # Behaviour is dependent on Neo4j version
    db_version = db.database_version
    if db_version.startswith("4"):
        # Nodes' ids
        assert lex_hives.id == int(lex_hives.element_id)
        assert lex_hives.id == (the_hives.released.single()).id
        # Relationships' ids
        assert isinstance(released_rel.element_id, str)
        assert int(released_rel.element_id) == released_rel.id
        assert released_rel._start_node_id == int(the_hives.element_id)
        assert released_rel._end_node_id == int(lex_hives.element_id)
    else:
        # Nodes' ids
        expected_error_type = ValueError
        expected_error_message = "id is deprecated in Neo4j version 5, please migrate to element_id\. If you use the id in a Cypher query, replace id\(\) by elementId\(\)\."
        assert isinstance(lex_hives.element_id, str)
        with pytest.raises(
            expected_error_type,
            match=expected_error_message,
        ):
            lex_hives.id

        # Relationships' ids
        assert isinstance(released_rel.element_id, str)
        assert isinstance(released_rel._start_node_element_id, str)
        assert isinstance(released_rel._end_node_element_id, str)
        with pytest.raises(
            expected_error_type,
            match=expected_error_message,
        ):
            released_rel.id
        with pytest.raises(
            expected_error_type,
            match=expected_error_message,
        ):
            released_rel._start_node_id
        with pytest.raises(
            expected_error_type,
            match=expected_error_message,
        ):
            released_rel._end_node_id



"""
Provides a test case for issue 374 - "Support for Point property type".

For more information please see: https://github.com/neo4j-contrib/neomodel/issues/374
"""

import random
from test._async_compat import mark_async_test

import neo4j.spatial
import pytest

import neomodel
import neomodel.contrib.spatial_properties

from .test_spatial_datatypes import (
    basic_type_assertions,
    check_and_skip_neo4j_least_version,
)


def test_spatial_point_property():
    """
    Tests that specific modes of instantiation fail as expected.

    :return:
    """

    # Neo4j versions lower than 3.4.0 do not support Point. In that case, skip the test.
    check_and_skip_neo4j_least_version(
        340, "This version does not support spatial data types."
    )

    with pytest.raises(ValueError, match=r"Invalid CRS\(None\)"):
        a_point_property = neomodel.contrib.spatial_properties.PointProperty()

    with pytest.raises(ValueError, match=r"Invalid CRS\(crs_isaak\)"):
        a_point_property = neomodel.contrib.spatial_properties.PointProperty(
            crs="crs_isaak"
        )

    with pytest.raises(TypeError, match="Invalid default value"):
        a_point_property = neomodel.contrib.spatial_properties.PointProperty(
            default=(0.0, 0.0), crs="cartesian"
        )


def test_inflate():
    """
    Tests that the marshalling from neo4j to neomodel data types works as expected.

    :return:
    """

    # Neo4j versions lower than 3.4.0 do not support Point. In that case, skip the test.
    check_and_skip_neo4j_least_version(
        340, "This version does not support spatial data types."
    )

    # The test is repeatable enough to try and standardise it. The same test is repeated with the assertions in
    # `basic_type_assertions` and different messages to be able to localise the exception.
    #
    # Array of points to inflate and messages when things go wrong
    values_from_db = [
        (
            neo4j.spatial.CartesianPoint((0.0, 0.0)),
            "Expected Neomodel 2d cartesian point when inflating 2d cartesian neo4j point",
        ),
        (
            neo4j.spatial.CartesianPoint((0.0, 0.0, 0.0)),
            "Expected Neomodel 3d cartesian point when inflating 3d cartesian neo4j point",
        ),
        (
            neo4j.spatial.WGS84Point((0.0, 0.0)),
            "Expected Neomodel 2d geographical point when inflating 2d geographical neo4j point",
        ),
        (
            neo4j.spatial.WGS84Point((0.0, 0.0, 0.0)),
            "Expected Neomodel 3d geographical point inflating 3d geographical neo4j point",
        ),
    ]

    # Run the above tests
    for a_value in values_from_db:
        expected_point = neomodel.contrib.spatial_properties.NeomodelPoint(
            tuple(a_value[0]),
            crs=neomodel.contrib.spatial_properties.SRID_TO_CRS[a_value[0].srid],
        )
        inflated_point = neomodel.contrib.spatial_properties.PointProperty(
            crs=neomodel.contrib.spatial_properties.SRID_TO_CRS[a_value[0].srid]
        ).inflate(a_value[0])
        basic_type_assertions(
            expected_point,
            inflated_point,
            "{}, received {}".format(a_value[1], inflated_point),
        )


def test_deflate():
    """
    Tests that the marshalling from neomodel to neo4j data types works as expected
    :return:
    """
    # Please see inline comments in `test_inflate`. This test function is 90% to that one with very minor differences.
    #

    # Neo4j versions lower than 3.4.0 do not support Point. In that case, skip the test.
    check_and_skip_neo4j_least_version(
        340, "This version does not support spatial data types."
    )

    CRS_TO_SRID = dict(
        [
            (value, key)
            for key, value in neomodel.contrib.spatial_properties.SRID_TO_CRS.items()
        ]
    )
    # Values to construct and expect during deflation
    values_from_neomodel = [
        (
            neomodel.contrib.spatial_properties.NeomodelPoint(
                (0.0, 0.0), crs="cartesian"
            ),
            "Expected Neo4J 2d cartesian point when deflating Neomodel 2d cartesian point",
        ),
        (
            neomodel.contrib.spatial_properties.NeomodelPoint(
                (0.0, 0.0, 0.0), crs="cartesian-3d"
            ),
            "Expected Neo4J 3d cartesian point when deflating Neomodel 3d cartesian point",
        ),
        (
            neomodel.contrib.spatial_properties.NeomodelPoint((0.0, 0.0), crs="wgs-84"),
            "Expected Neo4J 2d geographical point when deflating Neomodel 2d geographical point",
        ),
        (
            neomodel.contrib.spatial_properties.NeomodelPoint(
                (0.0, 0.0, 0.0), crs="wgs-84-3d"
            ),
            "Expected Neo4J 3d geographical point when deflating Neomodel 3d geographical point",
        ),
    ]

    # Run the above tests.
    for a_value in values_from_neomodel:
        expected_point = neo4j.spatial.Point(tuple(a_value[0].coords[0]))
        expected_point.srid = CRS_TO_SRID[a_value[0].crs]
        deflated_point = neomodel.contrib.spatial_properties.PointProperty(
            crs=a_value[0].crs
        ).deflate(a_value[0])
        basic_type_assertions(
            expected_point,
            deflated_point,
            "{}, received {}".format(a_value[1], deflated_point),
            check_neo4j_points=True,
        )


@mark_async_test
async def test_default_value():
    """
    Tests that the default value passing mechanism works as expected with NeomodelPoint values.
    :return:
    """

    def get_some_point():
        return neomodel.contrib.spatial_properties.NeomodelPoint(
            (random.random(), random.random())
        )

    class LocalisableEntity(neomodel.AsyncStructuredNode):
        """
        A very simple entity to try out the default value assignment.
        """

        identifier = neomodel.UniqueIdProperty()
        location = neomodel.contrib.spatial_properties.PointProperty(
            crs="cartesian", default=get_some_point
        )

    # Neo4j versions lower than 3.4.0 do not support Point. In that case, skip the test.
    check_and_skip_neo4j_least_version(
        340, "This version does not support spatial data types."
    )

    # Save an object
    an_object = await LocalisableEntity().save()
    coords = an_object.location.coords[0]
    # Retrieve it
    retrieved_object = await LocalisableEntity.nodes.get(
        identifier=an_object.identifier
    )
    # Check against an independently created value
    assert (
        retrieved_object.location
        == neomodel.contrib.spatial_properties.NeomodelPoint(coords)
    ), ("Default value assignment failed.")


@mark_async_test
async def test_array_of_points():
    """
    Tests that Arrays of Points work as expected.

    :return:
    """

    class AnotherLocalisableEntity(neomodel.AsyncStructuredNode):
        """
        A very simple entity with an array of locations
        """

        identifier = neomodel.UniqueIdProperty()
        locations = neomodel.ArrayProperty(
            neomodel.contrib.spatial_properties.PointProperty(crs="cartesian")
        )

    # Neo4j versions lower than 3.4.0 do not support Point. In that case, skip the test.
    check_and_skip_neo4j_least_version(
        340, "This version does not support spatial data types."
    )

    an_object = await AnotherLocalisableEntity(
        locations=[
            neomodel.contrib.spatial_properties.NeomodelPoint((0.0, 0.0)),
            neomodel.contrib.spatial_properties.NeomodelPoint((1.0, 0.0)),
        ]
    ).save()

    retrieved_object = await AnotherLocalisableEntity.nodes.get(
        identifier=an_object.identifier
    )

    assert (
        type(retrieved_object.locations) is list
    ), "Array of Points definition failed."
    assert retrieved_object.locations == [
        neomodel.contrib.spatial_properties.NeomodelPoint((0.0, 0.0)),
        neomodel.contrib.spatial_properties.NeomodelPoint((1.0, 0.0)),
    ], "Array of Points incorrect values."


@mark_async_test
async def test_simple_storage_retrieval():
    """
    Performs a simple Create, Retrieve via .save(), .get() which, due to the way Q objects operate, tests the
    __copy__, __deepcopy__ operations of NeomodelPoint.
    :return:
    """

    class TestStorageRetrievalProperty(neomodel.AsyncStructuredNode):
        uid = neomodel.UniqueIdProperty()
        description = neomodel.StringProperty()
        location = neomodel.contrib.spatial_properties.PointProperty(crs="cartesian")

    # Neo4j versions lower than 3.4.0 do not support Point. In that case, skip the test.
    check_and_skip_neo4j_least_version(
        340, "This version does not support spatial data types."
    )

    a_restaurant = await TestStorageRetrievalProperty(
        description="Milliways",
        location=neomodel.contrib.spatial_properties.NeomodelPoint((0, 0)),
    ).save()

    a_property = await TestStorageRetrievalProperty.nodes.get(
        location=neomodel.contrib.spatial_properties.NeomodelPoint((0, 0))
    )

    assert a_restaurant.description == a_property.description


def test_equality_with_other_objects():
    """
    Performs equality tests and ensures tha ``NeomodelPoint`` can be compared with ShapelyPoint and NeomodelPoint only.
    """
    try:
        import shapely.geometry
        from shapely import __version__
    except ImportError:
        pytest.skip("Shapely module not present")

    if int("".join(__version__.split(".")[0:3])) < 200:
        pytest.skip(f"Shapely 2.0 not present (Current version is {__version__}")

    assert neomodel.contrib.spatial_properties.NeomodelPoint(
        (0, 0)
    ) == neomodel.contrib.spatial_properties.NeomodelPoint(x=0, y=0)
    assert neomodel.contrib.spatial_properties.NeomodelPoint(
        (0, 0)
    ) == shapely.geometry.Point((0, 0))



"""
Provides integration with `pandas <https://pandas.pydata.org/>`_.

.. note::
   This module requires pandas to be installed, and will raise a
   warning if this is not available.

Example:

    >>> from neomodel.async_ import db
    >>> from neomodel.integration.pandas import to_dataframe
    >>> db.set_connection('bolt://neo4j:secret@localhost:7687')
    >>> df = to_dataframe(db.cypher_query("MATCH (u:User) RETURN u.email AS email, u.name AS name"))
    >>> df
                       email    name
    0         jimla@test.com   jimla
    1         jimlo@test.com   jimlo

    [2 rows x 2 columns]

"""


from warnings import warn

try:
    # noinspection PyPackageRequirements
    from pandas import DataFrame, Series
except ImportError:
    warn(
        "The neomodel.integration.pandas module expects pandas to be installed "
        "but it does not appear to be available."
    )
    raise


def to_dataframe(query_results: tuple, index=None, dtype=None):
    """Convert the results of a db.cypher_query call and associated metadata
    into a pandas DataFrame.
    Optionally, specify an index and/or a datatype for the columns.
    """
    results, meta = query_results
    return DataFrame(results, columns=meta, index=index, dtype=dtype)


def to_series(query_results: tuple, field=0, index=None, dtype=None):
    """Convert the results of a db.cypher_query call
    into a pandas Series for the given field.
    Optionally, specify an index and/or a datatype for the columns.
    """
    results, _ = query_results
    return Series([record[field] for record in results], index=index, dtype=dtype)



from datetime import date, datetime, timedelta
from test._async_compat import mark_async_test

from neo4j import time
from pytest import mark, raises
from pytz import timezone

from neomodel import (
    AsyncRelationship,
    AsyncStructuredNode,
    AsyncStructuredRel,
    adb,
    config,
)
from neomodel.contrib import AsyncSemiStructuredNode
from neomodel.exceptions import (
    DeflateError,
    InflateError,
    RequiredProperty,
    UniqueProperty,
)
from neomodel.properties import (
    AliasProperty,
    ArrayProperty,
    BooleanProperty,
    DateProperty,
    DateTimeFormatProperty,
    DateTimeNeo4jFormatProperty,
    DateTimeProperty,
    EmailProperty,
    IntegerProperty,
    JSONProperty,
    NormalizedProperty,
    RegexProperty,
    StringProperty,
    UniqueIdProperty,
    validator,
)
from neomodel.util import get_graph_entity_properties


class FooBar:
    pass


def test_string_property_exceeds_max_length():
    """
    StringProperty is defined by two properties: `max_length` and `choices` that are mutually exclusive. Furthermore,
    max_length must be a positive non-zero number.
    """
    # Try to define a property that has both choices and max_length
    with raises(ValueError):
        some_string_property = StringProperty(
            choices={"One": "1", "Two": "2"}, max_length=22
        )

    # Try to define a string property that has a negative zero length
    with raises(ValueError):
        another_string_property = StringProperty(max_length=-35)

    # Try to validate a long string
    a_string_property = StringProperty(required=True, max_length=5)
    with raises(ValueError):
        a_string_property.normalize("The quick brown fox jumps over the lazy dog")

    # Try to validate a "valid" string, as per the max_length setting.
    valid_string = "Owen"
    normalised_string = a_string_property.normalize(valid_string)
    assert (
        valid_string == normalised_string
    ), "StringProperty max_length test passed but values do not match."


@mark_async_test
async def test_string_property_w_choice():
    class TestChoices(AsyncStructuredNode):
        SEXES = {"F": "Female", "M": "Male", "O": "Other"}
        sex = StringProperty(required=True, choices=SEXES)

    try:
        await TestChoices(sex="Z").save()
    except DeflateError as e:
        assert "choice" in str(e)
    else:
        assert False, "DeflateError not raised."

    node = await TestChoices(sex="M").save()
    assert node.get_sex_display() == "Male"

    with raises(ValueError):

        class WrongChoices(AsyncStructuredNode):
            WRONG = "wrong"
            wrong_prop = StringProperty(choices=WRONG)


def test_deflate_inflate():
    prop = IntegerProperty(required=True)
    prop.name = "age"
    prop.owner = FooBar

    try:
        prop.inflate("six")
    except InflateError as e:
        assert "inflate property" in str(e)
    else:
        assert False, "DeflateError not raised."

    try:
        prop.deflate("six")
    except DeflateError as e:
        assert "deflate property" in str(e)
    else:
        assert False, "DeflateError not raised."

    with raises(ValueError, match="Unknown Property method tartiflate"):

        class CheeseProperty(IntegerProperty):
            @validator
            def tartiflate(self, value):
                return int(value)


def test_boolean_property():
    prop = BooleanProperty(default=False)
    prop.name = "foo"
    prop.owner = FooBar
    assert prop.deflate(True) is True
    assert prop.deflate(False) is False
    assert prop.inflate(True) is True
    assert prop.inflate(False) is False

    assert prop.default_value() is False


def test_datetimes_timezones():
    prop = DateTimeProperty()
    prop.name = "foo"
    prop.owner = FooBar
    t = datetime.utcnow()
    gr = timezone("Europe/Athens")
    gb = timezone("Europe/London")
    dt1 = gr.localize(t)
    dt2 = gb.localize(t)
    time1 = prop.inflate(prop.deflate(dt1))
    time2 = prop.inflate(prop.deflate(dt2))
    assert time1.utctimetuple() == dt1.utctimetuple()
    assert time1.utctimetuple() < time2.utctimetuple()
    assert time1.tzname() == "UTC"

    with raises(ValueError, match="too many defaults"):
        _ = DateTimeFormatProperty(
            default_now=True, default=datetime(1900, 1, 1, 0, 0, 0)
        )

    prev_force_timezone = config.FORCE_TIMEZONE
    config.FORCE_TIMEZONE = True
    with raises(ValueError, match=r".*No timezone provided."):
        prop.deflate(datetime.now())

    config.FORCE_TIMEZONE = prev_force_timezone


def test_date():
    prop = DateProperty()
    prop.name = "foo"
    prop.owner = FooBar
    somedate = date(2012, 12, 15)
    assert prop.deflate(somedate) == "2012-12-15"
    assert prop.inflate("2012-12-15") == somedate

    assert prop.inflate(time.DateTime(2007, 9, 27)) == date(2007, 9, 27)


def test_datetime_format():
    some_format = "%Y-%m-%d %H:%M:%S"
    prop = DateTimeFormatProperty(format=some_format)
    prop.name = "foo"
    prop.owner = FooBar
    some_datetime = datetime(2019, 3, 19, 15, 36, 25)
    assert prop.deflate(some_datetime) == "2019-03-19 15:36:25"
    assert prop.inflate("2019-03-19 15:36:25") == some_datetime

    with raises(ValueError, match=r"datetime object expected, got.*"):
        prop.deflate(1234)

    with raises(ValueError, match="too many defaults"):
        _ = DateTimeFormatProperty(
            default_now=True, default=datetime(1900, 1, 1, 0, 0, 0)
        )

    secondProp = DateTimeFormatProperty(default_now=True)
    assert secondProp.has_default
    assert (
        timedelta(seconds=-2)
        < secondProp.default - datetime.now()
        < timedelta(seconds=2)
    )


def test_datetime_neo4j_format():
    prop = DateTimeNeo4jFormatProperty()
    prop.name = "foo"
    prop.owner = FooBar
    some_datetime = datetime(2022, 12, 10, 14, 00, 00)
    assert prop.has_default is False
    assert prop.default is None
    assert prop.deflate(some_datetime) == time.DateTime(2022, 12, 10, 14, 00, 00)
    assert prop.inflate(time.DateTime(2022, 12, 10, 14, 00, 00)) == some_datetime

    with raises(ValueError, match=r"datetime object expected, got.*"):
        prop.deflate(1234)

    with raises(ValueError, match="too many defaults"):
        _ = DateTimeNeo4jFormatProperty(
            default_now=True, default=datetime(1900, 1, 1, 0, 0, 0)
        )

    secondProp = DateTimeNeo4jFormatProperty(default_now=True)
    assert secondProp.has_default
    assert (
        timedelta(seconds=-2)
        < secondProp.default - datetime.now()
        < timedelta(seconds=2)
    )


def test_datetime_exceptions():
    prop = DateTimeProperty()
    prop.name = "created"
    prop.owner = FooBar
    faulty = "dgdsg"

    try:
        prop.inflate(faulty)
    except InflateError as e:
        assert "inflate property" in str(e)
    else:
        assert False, "InflateError not raised."

    try:
        prop.deflate(faulty)
    except DeflateError as e:
        assert "deflate property" in str(e)
    else:
        assert False, "DeflateError not raised."

    with raises(ValueError, match="too many defaults"):
        _ = DateTimeProperty(default_now=True, default=datetime(1900, 1, 1, 0, 0, 0))


def test_date_exceptions():
    prop = DateProperty()
    prop.name = "date"
    prop.owner = FooBar
    faulty = "2012-14-13"

    try:
        prop.inflate(faulty)
    except InflateError as e:
        assert "inflate property" in str(e)
    else:
        assert False, "InflateError not raised."

    try:
        prop.deflate(faulty)
    except DeflateError as e:
        assert "deflate property" in str(e)
    else:
        assert False, "DeflateError not raised."


def test_base_exceptions():
    # default-required conflict
    with raises(
        ValueError,
        match="The arguments `required` and `default` are mutually exclusive.",
    ):
        _ = StringProperty(default="kakapo", required=True)

    # unique_index - index conflict
    with raises(
        ValueError,
        match="The arguments `unique_index` and `index` are mutually exclusive.",
    ):
        _ = IntegerProperty(index=True, unique_index=True)

    # no default value
    kakapo = StringProperty()
    with raises(ValueError, match="No default value specified"):
        kakapo.default_value()

    # missing normalize method
    class WoopsProperty(NormalizedProperty):
        pass

    woops = WoopsProperty()
    with raises(NotImplementedError, match="Specialize normalize method"):
        woops.normalize("kakapo")


def test_json():
    prop = JSONProperty()
    prop.name = "json"
    prop.owner = FooBar

    value = {"test": [1, 2, 3]}

    assert prop.deflate(value) == '{"test": [1, 2, 3]}'
    assert prop.inflate('{"test": [1, 2, 3]}') == value

    value_with_unicode = {"test": [1, 2, 3, "©"]}
    assert prop.deflate(value_with_unicode) == '{"test": [1, 2, 3, "\\u00a9"]}'
    assert prop.inflate('{"test": [1, 2, 3, "\\u00a9"]}') == value_with_unicode


def test_json_unicode():
    prop = JSONProperty(ensure_ascii=False)
    prop.name = "json"
    prop.owner = FooBar

    value = {"test": [1, 2, 3, "©"]}

    assert prop.deflate(value) == '{"test": [1, 2, 3, "©"]}'
    assert prop.inflate('{"test": [1, 2, 3, "©"]}') == value


def test_indexed():
    indexed = StringProperty(index=True)
    assert indexed.is_indexed is True

    unique_indexed = StringProperty(unique_index=True)
    assert unique_indexed.is_indexed is True

    not_indexed = StringProperty()
    assert not_indexed.is_indexed is False


@mark_async_test
async def test_default_value():
    class DefaultTestValue(AsyncStructuredNode):
        name_xx = StringProperty(default="jim", index=True)

    a = DefaultTestValue()
    assert a.name_xx == "jim"
    await a.save()


@mark_async_test
async def test_default_value_callable():
    def uid_generator():
        return "xx"

    class DefaultTestValueTwo(AsyncStructuredNode):
        uid = StringProperty(default=uid_generator, index=True)

    a = await DefaultTestValueTwo().save()
    assert a.uid == "xx"


@mark_async_test
async def test_default_value_callable_type():
    # check our object gets converted to str without serializing and reload
    def factory():
        class Foo:
            def __str__(self):
                return "123"

        return Foo()

    class DefaultTestValueThree(AsyncStructuredNode):
        uid = StringProperty(default=factory, index=True)

    x = DefaultTestValueThree()
    assert x.uid == "123"
    await x.save()
    assert x.uid == "123"
    await x.refresh()
    assert x.uid == "123"


class TestDBNamePropertyRel(AsyncStructuredRel):
    known_for = StringProperty(db_property="knownFor")


# This must be defined outside of the test, otherwise the `Relationship` definition cannot look up
# `TestDBNamePropertyNode`
class TestDBNamePropertyNode(AsyncStructuredNode):
    name_ = StringProperty(db_property="name")
    knows = AsyncRelationship(
        "TestDBNamePropertyNode", "KNOWS", model=TestDBNamePropertyRel
    )


@mark_async_test
async def test_independent_property_name():
    # -- test node --
    x = TestDBNamePropertyNode()
    x.name_ = "jim"
    await x.save()

    # check database property name on low level
    results, meta = await adb.cypher_query("MATCH (n:TestDBNamePropertyNode) RETURN n")
    node_properties = get_graph_entity_properties(results[0][0])
    assert node_properties["name"] == "jim"
    assert "name_" not in node_properties

    # check python class property name at a high level
    assert not hasattr(x, "name")
    assert hasattr(x, "name_")
    assert (await TestDBNamePropertyNode.nodes.filter(name_="jim").all())[
        0
    ].name_ == x.name_
    assert (await TestDBNamePropertyNode.nodes.get(name_="jim")).name_ == x.name_

    # -- test relationship --

    r = await x.knows.connect(x)
    r.known_for = "10 years"
    await r.save()

    # check database property name on low level
    results, meta = await adb.cypher_query(
        "MATCH (:TestDBNamePropertyNode)-[r:KNOWS]->(:TestDBNamePropertyNode) RETURN r"
    )
    rel_properties = get_graph_entity_properties(results[0][0])
    assert rel_properties["knownFor"] == "10 years"
    assert not "known_for" in node_properties

    # check python class property name at a high level
    assert not hasattr(r, "knownFor")
    assert hasattr(r, "known_for")
    rel = await x.knows.relationship(x)
    assert rel.known_for == r.known_for


@mark_async_test
async def test_independent_property_name_for_semi_structured():
    class TestDBNamePropertySemiStructuredNode(AsyncSemiStructuredNode):
        title_ = StringProperty(db_property="title")

    semi = TestDBNamePropertySemiStructuredNode(title_="sir", extra="data")
    await semi.save()

    # check database property name on low level
    results, meta = await adb.cypher_query(
        "MATCH (n:TestDBNamePropertySemiStructuredNode) RETURN n"
    )
    node_properties = get_graph_entity_properties(results[0][0])
    assert node_properties["title"] == "sir"
    # assert "title_" not in node_properties
    assert node_properties["extra"] == "data"

    # check python class property name at a high level
    assert hasattr(semi, "title_")
    assert not hasattr(semi, "title")
    assert hasattr(semi, "extra")
    from_filter = (
        await TestDBNamePropertySemiStructuredNode.nodes.filter(title_="sir").all()
    )[0]
    assert from_filter.title_ == "sir"
    # assert not hasattr(from_filter, "title")
    assert from_filter.extra == "data"
    from_get = await TestDBNamePropertySemiStructuredNode.nodes.get(title_="sir")
    assert from_get.title_ == "sir"
    # assert not hasattr(from_get, "title")
    assert from_get.extra == "data"


@mark_async_test
async def test_independent_property_name_get_or_create():
    class TestNode(AsyncStructuredNode):
        uid = UniqueIdProperty()
        name_ = StringProperty(db_property="name", required=True)

    # create the node
    await TestNode.get_or_create({"uid": 123, "name_": "jim"})
    # test that the node is retrieved correctly
    x = (await TestNode.get_or_create({"uid": 123, "name_": "jim"}))[0]

    # check database property name on low level
    results, _ = await adb.cypher_query("MATCH (n:TestNode) RETURN n")
    node_properties = get_graph_entity_properties(results[0][0])
    assert node_properties["name"] == "jim"
    assert "name_" not in node_properties


@mark.parametrize("normalized_class", (NormalizedProperty,))
def test_normalized_property(normalized_class):
    class TestProperty(normalized_class):
        def normalize(self, value):
            self._called_with = value
            self._called = True
            return value + "bar"

    inflate = TestProperty()
    inflate_res = inflate.inflate("foo")
    assert getattr(inflate, "_called", False)
    assert getattr(inflate, "_called_with", None) == "foo"
    assert inflate_res == "foobar"

    deflate = TestProperty()
    deflate_res = deflate.deflate("bar")
    assert getattr(deflate, "_called", False)
    assert getattr(deflate, "_called_with", None) == "bar"
    assert deflate_res == "barbar"

    default = TestProperty(default="qux")
    default_res = default.default_value()
    assert getattr(default, "_called", False)
    assert getattr(default, "_called_with", None) == "qux"
    assert default_res == "quxbar"


def test_regex_property():
    class MissingExpression(RegexProperty):
        pass

    with raises(AttributeError):
        MissingExpression()

    class TestProperty(RegexProperty):
        name = "test"
        owner = object()
        expression = r"\w+ \w+$"

        def normalize(self, value):
            self._called = True
            return super().normalize(value)

    prop = TestProperty()
    result = prop.inflate("foo bar")
    assert getattr(prop, "_called", False)
    assert result == "foo bar"

    with raises(DeflateError):
        prop.deflate("qux")


def test_email_property():
    prop = EmailProperty()
    prop.name = "email"
    prop.owner = object()
    result = prop.inflate("foo@example.com")
    assert result == "foo@example.com"

    with raises(DeflateError):
        prop.deflate("foo@example")


@mark_async_test
async def test_uid_property():
    prop = UniqueIdProperty()
    prop.name = "uid"
    prop.owner = object()
    myuid = prop.default_value()
    assert len(myuid)

    class CheckMyId(AsyncStructuredNode):
        uid = UniqueIdProperty()

    cmid = await CheckMyId().save()
    assert len(cmid.uid)

    matched_exception = r".*argument ignored by.*"
    # Test ignored arguments
    with raises(ValueError, match=matched_exception):
        _ = UniqueIdProperty(required=False)

    with raises(ValueError, match=matched_exception):
        _ = UniqueIdProperty(unique_index=False)

    with raises(ValueError, match=matched_exception):
        _ = UniqueIdProperty(index=False)

    with raises(ValueError, match=matched_exception):
        _ = UniqueIdProperty(default="kakapo")


class ArrayProps(AsyncStructuredNode):
    uid = StringProperty(unique_index=True)
    untyped_arr = ArrayProperty()
    typed_arr = ArrayProperty(IntegerProperty())


@mark_async_test
async def test_array_properties():
    # untyped
    ap1 = await ArrayProps(uid="1", untyped_arr=["Tim", "Bob"]).save()
    assert "Tim" in ap1.untyped_arr
    ap1 = await ArrayProps.nodes.get(uid="1")
    assert "Tim" in ap1.untyped_arr

    # typed
    try:
        await ArrayProps(uid="2", typed_arr=["a", "b"]).save()
    except DeflateError as e:
        assert "unsaved node" in str(e)
    else:
        assert False, "DeflateError not raised."

    ap2 = await ArrayProps(uid="2", typed_arr=[1, 2]).save()
    assert 1 in ap2.typed_arr
    ap2 = await ArrayProps.nodes.get(uid="2")
    assert 2 in ap2.typed_arr

    class Kakapo:
        pass

    with raises(TypeError, match="Expecting neomodel Property"):
        ArrayProperty(Kakapo)

    with raises(TypeError, match="Cannot have nested ArrayProperty"):
        ArrayProperty(ArrayProperty())


def test_illegal_array_base_prop_raises():
    with raises(ValueError):
        ArrayProperty(StringProperty(index=True))


@mark_async_test
async def test_indexed_array():
    class IndexArray(AsyncStructuredNode):
        ai = ArrayProperty(unique_index=True)

    b = await IndexArray(ai=[1, 2]).save()
    c = await IndexArray.nodes.get(ai=[1, 2])
    assert b.element_id == c.element_id


@mark_async_test
async def test_unique_index_prop_not_required():
    class ConstrainedTestNode(AsyncStructuredNode):
        required_property = StringProperty(required=True)
        unique_property = StringProperty(unique_index=True)
        unique_required_property = StringProperty(unique_index=True, required=True)
        unconstrained_property = StringProperty()

    # Create a node with a missing required property
    with raises(RequiredProperty):
        x = ConstrainedTestNode(required_property="required", unique_property="unique")
        await x.save()

    # Create a node with a missing unique (but not required) property.
    x = ConstrainedTestNode()
    x.required_property = "required"
    x.unique_required_property = "unique and required"
    x.unconstrained_property = "no contraints"
    await x.save()

    # check database property name on low level
    results, meta = await adb.cypher_query("MATCH (n:ConstrainedTestNode) RETURN n")
    node_properties = get_graph_entity_properties(results[0][0])
    assert node_properties["unique_required_property"] == "unique and required"


@mark_async_test
async def test_unique_index_prop_enforced():
    class UniqueNullableNameNode(AsyncStructuredNode):
        name = StringProperty(unique_index=True)

    await adb.install_labels(UniqueNullableNameNode)
    # Nameless
    x = UniqueNullableNameNode()
    await x.save()
    y = UniqueNullableNameNode()
    await y.save()

    # Named
    z = UniqueNullableNameNode(name="named")
    await z.save()
    with raises(UniqueProperty):
        a = UniqueNullableNameNode(name="named")
        await a.save()

    # Check nodes are in database
    results, _ = await adb.cypher_query("MATCH (n:UniqueNullableNameNode) RETURN n")
    assert len(results) == 3


def test_alias_property():
    class AliasedClass(AsyncStructuredNode):
        name = StringProperty(index=True)
        national_id = IntegerProperty(unique_index=True)
        alias = AliasProperty(to="name")
        alias_national_id = AliasProperty(to="national_id")
        whatever = StringProperty()
        alias_whatever = AliasProperty(to="whatever")

    assert AliasedClass.alias.index is True
    assert AliasedClass.alias_national_id.unique_index is True
    assert AliasedClass.alias_whatever.index is False



from test._async_compat import mark_sync_test

from neomodel import AliasProperty, StringProperty, StructuredNode


class MagicProperty(AliasProperty):
    def setup(self):
        self.owner.setup_hook_called = True


class AliasTestNode(StructuredNode):
    name = StringProperty(unique_index=True)
    full_name = AliasProperty(to="name")
    long_name = MagicProperty(to="name")


@mark_sync_test
def test_property_setup_hook():
    timmy = AliasTestNode(long_name="timmy").save()
    assert AliasTestNode.setup_hook_called
    assert timmy.name == "timmy"


@mark_sync_test
def test_alias():
    jim = AliasTestNode(full_name="Jim").save()
    assert jim.name == "Jim"
    assert jim.full_name == "Jim"
    assert "full_name" not in AliasTestNode.deflate(jim.__properties__)
    jim = AliasTestNode.nodes.get(full_name="Jim")
    assert jim
    assert jim.name == "Jim"
    assert jim.full_name == "Jim"
    assert "full_name" not in AliasTestNode.deflate(jim.__properties__)



from datetime import datetime
from test._async_compat import mark_sync_test

import pytest

from neomodel import (
    ArrayProperty,
    DateTimeProperty,
    FloatProperty,
    RelationshipFrom,
    StringProperty,
    StructuredNode,
    StructuredRel,
    VectorIndex,
    db,
)
from neomodel.semantic_filters import VectorFilter


@mark_sync_test
def test_base_vectorfilter_async():
    """
    Tests that the vectorquery is run, node and score are returned.
    """

    # Vector Indexes only exist from 5.13 onwards
    if not db.version_is_higher_than("5.13"):
        pytest.skip("Vector Index not Generally Available in Neo4j.")

    class someNode(StructuredNode):
        name = StringProperty()
        vector = ArrayProperty(
            base_property=FloatProperty(), vector_index=VectorIndex(2, "cosine")
        )

    db.install_labels(someNode)

    john = someNode(name="John", vector=[float(0.5), float(0.5)]).save()
    fred = someNode(name="Fred", vector=[float(1.0), float(0.0)]).save()

    someNodeSearch = someNode.nodes.filter(
        vector_filter=VectorFilter(
            topk=3, vector_attribute_name="vector", candidate_vector=[0.25, 0]
        )
    )
    result = someNodeSearch.all()
    assert all(isinstance(x[0], someNode) for x in result)
    assert all(isinstance(x[1], float) for x in result)


@mark_sync_test
def test_vectorfilter_with_node_propertyfilter():
    """
    Tests that the vector query is run, and "john" node is the only node returned.
    """
    # Vector Indexes only exist from 5.13 onwards
    if not db.version_is_higher_than("5.13"):
        pytest.skip("Vector Index not Generally Available in Neo4j.")

    class someNodeBis(StructuredNode):
        name = StringProperty()
        vector = ArrayProperty(
            base_property=FloatProperty(), vector_index=VectorIndex(2, "cosine")
        )

    db.install_labels(someNodeBis)

    john = someNodeBis(name="John", vector=[float(0.5), float(0.5)]).save()
    fred = someNodeBis(name="Fred", vector=[float(1.0), float(0.0)]).save()

    vectorsearchFilterforJohn = someNodeBis.nodes.filter(
        vector_filter=VectorFilter(
            topk=3, vector_attribute_name="vector", candidate_vector=[0.25, 0]
        ),
        name="John",
    )
    result = vectorsearchFilterforJohn.all()

    assert len(result) == 1
    assert all(isinstance(x[0], someNodeBis) for x in result)
    assert result[0][0].name == "John"
    assert all(isinstance(x[1], float) for x in result)


@mark_sync_test
def test_dont_duplicate_vector_filter_node():
    """
    Tests the situation that another node have the same filter value.
    Testing that we are only perfomring the vectorfilter and metadata filter on the right nodes.
    """
    # Vector Indexes only exist from 5.13 onwards
    if not db.version_is_higher_than("5.13"):
        pytest.skip("Vector Index not Generally Available in Neo4j.")

    class someNodeTer(StructuredNode):
        name = StringProperty()
        vector = ArrayProperty(
            base_property=FloatProperty(), vector_index=VectorIndex(2, "cosine")
        )

    class otherNode(StructuredNode):
        otherName = StringProperty()
        other_vector = ArrayProperty(
            base_property=FloatProperty(), vector_index=VectorIndex(2, "cosine")
        )

    db.install_labels(someNodeTer)
    db.install_labels(otherNode)

    john = someNodeTer(name="John", vector=[float(0.5), float(0.5)]).save()
    fred = someNodeTer(name="Fred", vector=[float(1.0), float(0.0)]).save()
    john2 = otherNode(name="John", vector=[float(0.5), float(0.1)]).save()
    fred2 = otherNode(name="Fred", vector=[float(0.9), float(0.2)]).save()

    john_vector_search = someNodeTer.nodes.filter(
        vector_filter=VectorFilter(
            topk=3, vector_attribute_name="vector", candidate_vector=[0.25, 0]
        ),
        name="John",
    )
    result = john_vector_search.all()

    assert len(result) == 1  # Check we only get the one John
    assert isinstance(result[0][0], someNodeTer)  # check we only get the someNode John
    assert result[0][0].name == "John"
    assert isinstance(result[0][1], float)


@mark_sync_test
def test_django_filter_w_vector_filter():
    """
    Tests that django filters still work with the vector filter on.
    """
    # Vector Indexes only exist from 5.13 onwards
    if not db.version_is_higher_than("5.13"):
        pytest.skip("Vector Index not Generally Available in Neo4j.")

    class djangoNode(StructuredNode):
        name = StringProperty()
        vector = ArrayProperty(
            base_property=FloatProperty(), vector_index=VectorIndex(2, "cosine")
        )
        number = FloatProperty()

    db.install_labels(djangoNode)

    nodeone = djangoNode(
        name="John", vector=[float(0.5), float(0.5)], number=float(10)
    ).save()
    nodetwo = djangoNode(
        name="Fred", vector=[float(0.8), float(0.5)], number=float(3)
    ).save()

    vector_search_with_django_filter = djangoNode.nodes.filter(
        vector_filter=VectorFilter(
            topk=10, vector_attribute_name="vector", candidate_vector=[0.25, 0.25]
        ),
        number__gt=5,
    )
    result = vector_search_with_django_filter.all()
    assert len(result) == 1  # we only get the one node
    assert isinstance(result[0][0], djangoNode)
    assert result[0][0].number > 5


@mark_sync_test
def test_vectorfilter_with_relationshipfilter():
    """
    Tests that by filtering on a vector similarity and then performing a relationshipfilter
    """
    # Vector Indexes only exist from 5.13 onwards
    if not db.version_is_higher_than("5.13"):
        pytest.skip("Vector Index not Generally Available in Neo4j.")

    class SupplierV(StructuredNode):
        name = StringProperty()

    class SuppliesVRel(StructuredRel):
        since = DateTimeProperty(default=datetime.now)

    class ProductV(StructuredNode):
        name = StringProperty()
        description = StringProperty()
        description_embedding = ArrayProperty(
            FloatProperty(), vector_index=VectorIndex(dimensions=2)
        )
        suppliers = RelationshipFrom(SupplierV, "SUPPLIES", model=SuppliesVRel)

    db.install_labels(SupplierV)
    db.install_labels(SuppliesVRel)
    db.install_labels(ProductV)

    supplier1 = SupplierV(name="Supplier 1").save()
    supplier2 = SupplierV(name="Supplier 2").save()
    product1 = ProductV(
        name="Product A",
        description="High quality product",
        description_embedding=[0.1, 0.2],
    ).save()
    product2 = ProductV(
        name="Product B",
        description="High quality product",
        description_embedding=[0.2, 0.2],
    ).save()
    product1.suppliers.connect(supplier1)
    product1.suppliers.connect(supplier2)
    product2.suppliers.connect(supplier1)

    filtered_product = ProductV.nodes.filter(
        vector_filter=VectorFilter(
            topk=1,
            vector_attribute_name="description_embedding",
            candidate_vector=[0.1, 0.1],
        ),
        suppliers__name="Supplier 1",
    )
    result = filtered_product.all()

    assert len(result) == 1
    assert isinstance(result[0][0], ProductV)
    assert isinstance(result[0][1], SupplierV)
    assert isinstance(result[0][2], SuppliesVRel)



from multiprocessing.pool import ThreadPool as Pool
from test._async_compat import mark_async_test

from neomodel import AsyncStructuredNode, StringProperty, adb


class ThingyMaBob(AsyncStructuredNode):
    name = StringProperty(unique_index=True, required=True)


async def thing_create(name):
    name = str(name)
    (thing,) = await ThingyMaBob.get_or_create({"name": name})
    return thing.name, name


@mark_async_test
async def test_concurrency():
    with Pool(5) as p:
        results = p.map(thing_create, range(50))
        for to_unpack in results:
            returned, sent = await to_unpack
            assert returned == sent
        await adb.close_connection()



from datetime import datetime
from test._async_compat import mark_async_test

import pytz
from pytest import raises

from neomodel import (
    AsyncRelationship,
    AsyncRelationshipTo,
    AsyncStructuredNode,
    AsyncStructuredRel,
    DateTimeProperty,
    DeflateError,
    StringProperty,
)
from neomodel._async_compat.util import AsyncUtil

HOOKS_CALLED = {"pre_save": 0, "post_save": 0}


class FriendRel(AsyncStructuredRel):
    since = DateTimeProperty(default=lambda: datetime.now(pytz.utc))


class HatesRel(FriendRel):
    reason = StringProperty()

    def pre_save(self):
        HOOKS_CALLED["pre_save"] += 1

    def post_save(self):
        HOOKS_CALLED["post_save"] += 1


class Badger(AsyncStructuredNode):
    name = StringProperty(unique_index=True)
    friend = AsyncRelationship("Badger", "FRIEND", model=FriendRel)
    hates = AsyncRelationshipTo("Stoat", "HATES", model=HatesRel)


class Stoat(AsyncStructuredNode):
    name = StringProperty(unique_index=True)
    hates = AsyncRelationshipTo("Badger", "HATES", model=HatesRel)


@mark_async_test
async def test_either_connect_with_rel_model():
    paul = await Badger(name="Paul").save()
    tom = await Badger(name="Tom").save()

    # creating rels
    new_rel = await tom.friend.disconnect(paul)
    new_rel = await tom.friend.connect(paul)
    assert isinstance(new_rel, FriendRel)
    assert isinstance(new_rel.since, datetime)

    # updating properties
    new_rel.since = datetime.now(pytz.utc)
    assert isinstance(await new_rel.save(), FriendRel)

    # start and end nodes are the opposite of what you'd expect when using either..
    # I've tried everything possible to correct this to no avail
    paul = await new_rel.start_node()
    tom = await new_rel.end_node()
    assert paul.name == "Tom"
    assert tom.name == "Paul"


@mark_async_test
async def test_direction_connect_with_rel_model():
    paul = await Badger(name="Paul the badger").save()
    ian = await Stoat(name="Ian the stoat").save()

    rel = await ian.hates.connect(
        paul, {"reason": "thinks paul should bath more often"}
    )
    assert isinstance(rel.since, datetime)
    assert isinstance(rel, FriendRel)
    assert rel.reason.startswith("thinks")
    rel.reason = "he smells"
    await rel.save()

    ian = await rel.start_node()
    assert isinstance(ian, Stoat)
    paul = await rel.end_node()
    assert isinstance(paul, Badger)

    assert ian.name.startswith("Ian")
    assert paul.name.startswith("Paul")

    rel = await ian.hates.relationship(paul)
    assert isinstance(rel, HatesRel)
    assert isinstance(rel.since, datetime)
    await rel.save()

    # test deflate checking
    rel.since = "2:30pm"
    with raises(DeflateError):
        await rel.save()

    # check deflate check via connect
    with raises(DeflateError):
        await paul.hates.connect(
            ian,
            {
                "reason": "thinks paul should bath more often",
                "since": "2:30pm",
            },
        )


@mark_async_test
async def test_traversal_where_clause():
    phill = await Badger(name="Phill the badger").save()
    tim = await Badger(name="Tim the badger").save()
    bob = await Badger(name="Bob the badger").save()
    rel = await tim.friend.connect(bob)
    now = datetime.now(pytz.utc)
    assert rel.since < now
    rel2 = await tim.friend.connect(phill)
    assert rel2.since > now
    friends = tim.friend.match(since__gt=now)
    assert len(await friends.all()) == 1


@mark_async_test
async def test_multiple_rels_exist_issue_223():
    # check a badger can dislike a stoat for multiple reasons
    phill = await Badger(name="Phill").save()
    ian = await Stoat(name="Stoat").save()

    rel_a = await phill.hates.connect(ian, {"reason": "a"})
    rel_b = await phill.hates.connect(ian, {"reason": "b"})
    assert rel_a.element_id != rel_b.element_id

    if AsyncUtil.is_async_code:
        ian_a = (await phill.hates.match(reason="a"))[0]
        ian_b = (await phill.hates.match(reason="b"))[0]
    else:
        ian_a = phill.hates.match(reason="a")[0]
        ian_b = phill.hates.match(reason="b")[0]
    assert ian_a.element_id == ian_b.element_id


@mark_async_test
async def test_retrieve_all_rels():
    tom = await Badger(name="tom").save()
    ian = await Stoat(name="ian").save()

    rel_a = await tom.hates.connect(ian, {"reason": "a"})
    rel_b = await tom.hates.connect(ian, {"reason": "b"})

    rels = await tom.hates.all_relationships(ian)
    assert len(rels) == 2
    assert rels[0].element_id in [rel_a.element_id, rel_b.element_id]
    assert rels[1].element_id in [rel_a.element_id, rel_b.element_id]


@mark_async_test
async def test_save_hook_on_rel_model():
    HOOKS_CALLED["pre_save"] = 0
    HOOKS_CALLED["post_save"] = 0

    paul = await Badger(name="PaulB").save()
    ian = await Stoat(name="IanS").save()

    rel = await ian.hates.connect(paul, {"reason": "yadda yadda"})
    await rel.save()

    assert HOOKS_CALLED["pre_save"] == 2
    assert HOOKS_CALLED["post_save"] == 2



#!/usr/bin/env python
"""
.. _neomodel_install_labels:

``neomodel_install_labels``
---------------------------

::

    usage: neomodel_install_labels [-h] [--db bolt://neo4j:neo4j@localhost:7687] <someapp.models/app.py> [<someapp.models/app.py> ...]
    
    Setup indexes and constraints on labels in Neo4j for your neomodel schema.
    
    If a connection URL is not specified, the tool will look up the environment 
    variable NEO4J_BOLT_URL. If that environment variable is not set, the tool
    will attempt to connect to the default URL bolt://neo4j:neo4j@localhost:7687

    Note : this script only has a synchronous mode.
    
    positional arguments:
      <someapp.models/app.py>
                            python modules or files with neomodel schema declarations.
    
    options:
      -h, --help            show this help message and exit
      --db bolt://neo4j:neo4j@localhost:7687
                            Neo4j Server URL
"""

import textwrap
from argparse import ArgumentParser, RawDescriptionHelpFormatter
from os import environ

from neomodel.scripts.utils import load_python_module_or_file
from neomodel.sync_.core import db


def main():
    parser = ArgumentParser(
        formatter_class=RawDescriptionHelpFormatter,
        description=textwrap.dedent(
            """
                                    Setup indexes and constraints on labels in Neo4j for your neomodel schema.

                                    If a connection URL is not specified, the tool will look up the environment 
                                    variable NEO4J_BOLT_URL. If that environment variable is not set, the tool
                                    will attempt to connect to the default URL bolt://neo4j:neo4j@localhost:7687
                                    """
        ),
    )

    parser.add_argument(
        "apps",
        metavar="<someapp.models/app.py>",
        type=str,
        nargs="+",
        help="python modules or files with neomodel schema declarations.",
    )

    parser.add_argument(
        "--db",
        metavar="bolt://neo4j:neo4j@localhost:7687",
        dest="neo4j_bolt_url",
        type=str,
        default="",
        help="Neo4j Server URL",
    )

    args = parser.parse_args()

    bolt_url = args.neo4j_bolt_url
    if len(bolt_url) == 0:
        bolt_url = environ.get("NEO4J_BOLT_URL", "bolt://neo4j:neo4j@localhost:7687")

    for app in args.apps:
        load_python_module_or_file(app)

    # Connect after to override any code in the module that may set the connection
    print(f"Connecting to {bolt_url}")
    db.set_connection(url=bolt_url)

    db.install_all_labels()


if __name__ == "__main__":
    main()



import re
from datetime import datetime
from test._async_compat import mark_async_test

from pytest import raises, skip, warns

from neomodel import (
    INCOMING,
    ArrayProperty,
    AsyncRelationshipFrom,
    AsyncRelationshipTo,
    AsyncStructuredNode,
    AsyncStructuredRel,
    AsyncZeroOrOne,
    DateTimeProperty,
    IntegerProperty,
    Q,
    StringProperty,
    UniqueIdProperty,
    adb,
)
from neomodel._async_compat.util import AsyncUtil
from neomodel.async_.match import (
    AsyncNodeSet,
    AsyncQueryBuilder,
    AsyncTraversal,
    Collect,
    Last,
    NodeNameResolver,
    Optional,
    Path,
    RawCypher,
    RelationNameResolver,
    Size,
)
from neomodel.exceptions import MultipleNodesReturned, RelationshipClassNotDefined


class SupplierRel(AsyncStructuredRel):
    since = DateTimeProperty(default=datetime.now)
    courier = StringProperty()


class Supplier(AsyncStructuredNode):
    name = StringProperty()
    delivery_cost = IntegerProperty()
    coffees = AsyncRelationshipTo("Coffee", "COFFEE SUPPLIERS", model=SupplierRel)


class Species(AsyncStructuredNode):
    name = StringProperty()
    tags = ArrayProperty(StringProperty(), default=list)
    coffees = AsyncRelationshipFrom(
        "Coffee", "COFFEE SPECIES", model=AsyncStructuredRel
    )


class Coffee(AsyncStructuredNode):
    name = StringProperty(unique_index=True)
    price = IntegerProperty()
    suppliers = AsyncRelationshipFrom(Supplier, "COFFEE SUPPLIERS", model=SupplierRel)
    species = AsyncRelationshipTo(Species, "COFFEE SPECIES", model=AsyncStructuredRel)
    id_ = IntegerProperty()


class Extension(AsyncStructuredNode):
    extension = AsyncRelationshipTo("Extension", "extension")


class CountryX(AsyncStructuredNode):
    code = StringProperty(unique_index=True, required=True)
    inhabitant = AsyncRelationshipFrom("PersonX", "IS_FROM")


class CityX(AsyncStructuredNode):
    name = StringProperty(required=True)
    country = AsyncRelationshipTo(CountryX, "FROM_COUNTRY")


class PersonX(AsyncStructuredNode):
    uid = UniqueIdProperty()
    name = StringProperty(unique_index=True)
    age = IntegerProperty(index=True, default=0)

    # traverse outgoing IS_FROM relations, inflate to Country objects
    country = AsyncRelationshipTo(CountryX, "IS_FROM")

    # traverse outgoing LIVES_IN relations, inflate to City objects
    city = AsyncRelationshipTo(CityX, "LIVES_IN")


class SoftwareDependency(AsyncStructuredNode):
    name = StringProperty(required=True)
    version = StringProperty(required=True)


class HasCourseRel(AsyncStructuredRel):
    level = StringProperty()
    start_date = DateTimeProperty()
    end_date = DateTimeProperty()


class Course(AsyncStructuredNode):
    name = StringProperty()


class Building(AsyncStructuredNode):
    name = StringProperty()


class Student(AsyncStructuredNode):
    name = StringProperty()

    parents = AsyncRelationshipTo("Student", "HAS_PARENT", model=AsyncStructuredRel)
    children = AsyncRelationshipFrom("Student", "HAS_PARENT", model=AsyncStructuredRel)
    lives_in = AsyncRelationshipTo(Building, "LIVES_IN", model=AsyncStructuredRel)
    courses = AsyncRelationshipTo(Course, "HAS_COURSE", model=HasCourseRel)
    preferred_course = AsyncRelationshipTo(
        Course,
        "HAS_PREFERRED_COURSE",
        model=AsyncStructuredRel,
        cardinality=AsyncZeroOrOne,
    )


@mark_async_test
async def test_filter_exclude_via_labels():
    await Coffee(name="Java", price=99).save()

    node_set = AsyncNodeSet(Coffee)
    qb = await AsyncQueryBuilder(node_set).build_ast()

    results = [node async for node in qb._execute()]

    assert "(coffee:Coffee)" in qb._ast.match
    assert qb._ast.result_class
    assert len(results) == 1
    assert isinstance(results[0], Coffee)
    assert results[0].name == "Java"

    # with filter and exclude
    await Coffee(name="Kenco", price=3).save()
    node_set = node_set.filter(price__gt=2).exclude(price__gt=6, name="Java")
    qb = await AsyncQueryBuilder(node_set).build_ast()

    results = [node async for node in qb._execute()]
    assert "(coffee:Coffee)" in qb._ast.match
    assert "NOT" in qb._ast.where[0]
    assert len(results) == 1
    assert results[0].name == "Kenco"


@mark_async_test
async def test_simple_has_via_label():
    nescafe = await Coffee(name="Nescafe", price=99).save()
    tesco = await Supplier(name="Tesco", delivery_cost=2).save()
    await nescafe.suppliers.connect(tesco)

    ns = AsyncNodeSet(Coffee).has(suppliers=True)
    qb = await AsyncQueryBuilder(ns).build_ast()
    results = [node async for node in qb._execute()]
    assert "COFFEE SUPPLIERS" in qb._ast.where[0]
    assert len(results) == 1
    assert results[0].name == "Nescafe"

    await Coffee(name="nespresso", price=99).save()
    ns = AsyncNodeSet(Coffee).has(suppliers=False)
    qb = await AsyncQueryBuilder(ns).build_ast()
    results = [node async for node in qb._execute()]
    assert len(results) > 0
    assert "NOT" in qb._ast.where[0]


@mark_async_test
async def test_get():
    await Coffee(name="1", price=3).save()
    assert await Coffee.nodes.get(name="1")

    with raises(Coffee.DoesNotExist):
        await Coffee.nodes.get(name="2")

    await Coffee(name="2", price=3).save()

    with raises(MultipleNodesReturned):
        await Coffee.nodes.get(price=3)


@mark_async_test
async def test_simple_traverse_with_filter():
    nescafe = await Coffee(name="Nescafe2", price=99).save()
    tesco = await Supplier(name="Tesco", delivery_cost=2).save()
    await nescafe.suppliers.connect(tesco)

    qb = AsyncQueryBuilder(
        AsyncNodeSet(source=nescafe).suppliers.match(since__lt=datetime.now())
    )

    _ast = await qb.build_ast()
    results = [node async for node in qb._execute()]

    assert qb._ast.lookup
    assert qb._ast.match
    assert qb._ast.return_clause.startswith("suppliers")
    assert len(results) == 1
    assert results[0].name == "Tesco"


@mark_async_test
async def test_double_traverse():
    nescafe = await Coffee(name="Nescafe plus", price=99).save()
    tesco = await Supplier(name="Asda", delivery_cost=2).save()
    await nescafe.suppliers.connect(tesco)
    await tesco.coffees.connect(await Coffee(name="Decafe", price=2).save())

    ns = AsyncNodeSet(AsyncNodeSet(source=nescafe).suppliers.match()).coffees.match()
    qb = await AsyncQueryBuilder(ns).build_ast()

    results = [node async for node in qb._execute()]
    assert len(results) == 2
    names = [n.name for n in results]
    assert "Decafe" in names
    assert "Nescafe plus" in names


@mark_async_test
async def test_count():
    await Coffee(name="Nescafe Gold", price=99).save()
    ast = await AsyncQueryBuilder(AsyncNodeSet(source=Coffee)).build_ast()
    count = await ast._count()
    assert count > 0

    await Coffee(name="Kawa", price=27).save()
    node_set = AsyncNodeSet(source=Coffee)
    node_set.skip = 1
    node_set.limit = 1
    ast = await AsyncQueryBuilder(node_set).build_ast()
    count = await ast._count()
    assert count == 1


@mark_async_test
async def test_len_and_iter_and_bool():
    iterations = 0

    await Coffee(name="Icelands finest").save()

    for c in await Coffee.nodes:
        iterations += 1
        await c.delete()

    assert iterations > 0

    assert len(await Coffee.nodes) == 0


@mark_async_test
async def test_slice():
    await Coffee(name="Icelands finest").save()
    await Coffee(name="Britains finest").save()
    await Coffee(name="Japans finest").save()

    # Branching tests because async needs extra brackets
    if AsyncUtil.is_async_code:
        assert len(list((await Coffee.nodes)[1:])) == 2
        assert len(list((await Coffee.nodes)[:1])) == 1
        assert isinstance((await Coffee.nodes)[1], Coffee)
        assert isinstance((await Coffee.nodes)[0], Coffee)
        assert len(list((await Coffee.nodes)[1:2])) == 1
    else:
        assert len(list(Coffee.nodes[1:])) == 2
        assert len(list(Coffee.nodes[:1])) == 1
        assert isinstance(Coffee.nodes[1], Coffee)
        assert isinstance(Coffee.nodes[0], Coffee)
        assert len(list(Coffee.nodes[1:2])) == 1


@mark_async_test
async def test_issue_208():
    # calls to match persist across queries.

    b = await Coffee(name="basics").save()
    l = await Supplier(name="lidl").save()
    a = await Supplier(name="aldi").save()

    await b.suppliers.connect(l, {"courier": "fedex"})
    await b.suppliers.connect(a, {"courier": "dhl"})

    assert len(await b.suppliers.match(courier="fedex"))
    assert len(await b.suppliers.match(courier="dhl"))


@mark_async_test
async def test_issue_589():
    node1 = await Extension().save()
    node2 = await Extension().save()
    assert node2 not in await node1.extension
    await node1.extension.connect(node2)
    assert node2 in await node1.extension


@mark_async_test
async def test_contains():
    expensive = await Coffee(price=1000, name="Pricey").save()
    asda = await Coffee(name="Asda", price=1).save()

    assert expensive in await Coffee.nodes.filter(price__gt=999)
    assert asda not in await Coffee.nodes.filter(price__gt=999)

    # bad value raises
    with raises(ValueError, match=r"Expecting StructuredNode instance"):
        if AsyncUtil.is_async_code:
            assert await Coffee.nodes.check_contains(2)
        else:
            assert 2 in Coffee.nodes

    # unsaved
    with raises(ValueError, match=r"Unsaved node"):
        if AsyncUtil.is_async_code:
            assert await Coffee.nodes.check_contains(Coffee())
        else:
            assert Coffee() in Coffee.nodes


@mark_async_test
async def test_order_by():
    c1 = await Coffee(name="Icelands finest", price=5).save()
    c2 = await Coffee(name="Britains finest", price=10).save()
    c3 = await Coffee(name="Japans finest", price=35).save()

    if AsyncUtil.is_async_code:
        assert ((await Coffee.nodes.order_by("price"))[0]).price == 5
        assert ((await Coffee.nodes.order_by("-price"))[0]).price == 35
    else:
        assert (Coffee.nodes.order_by("price")[0]).price == 5
        assert (Coffee.nodes.order_by("-price")[0]).price == 35

    ns = Coffee.nodes.order_by("-price")
    qb = await AsyncQueryBuilder(ns).build_ast()
    assert qb._ast.order_by
    ns = ns.order_by(None)
    qb = await AsyncQueryBuilder(ns).build_ast()
    assert not qb._ast.order_by
    ns = ns.order_by("?")
    qb = await AsyncQueryBuilder(ns).build_ast()
    assert qb._ast.with_clause == "coffee, rand() as r"
    assert qb._ast.order_by == ["r"]

    with raises(
        ValueError,
        match=r".*Neo4j internals like id or element_id are not allowed for use in this operation.",
    ):
        await Coffee.nodes.order_by("id").all()

    # Test order by on a relationship
    l = await Supplier(name="lidl2").save()
    await l.coffees.connect(c1)
    await l.coffees.connect(c2)
    await l.coffees.connect(c3)

    ordered_n = [n for n in await l.coffees.order_by("name")]
    assert ordered_n[0] == c2
    assert ordered_n[1] == c1
    assert ordered_n[2] == c3


@mark_async_test
async def test_order_by_rawcypher():
    d1 = await SoftwareDependency(name="Package1", version="1.0.0").save()
    d2 = await SoftwareDependency(name="Package2", version="1.4.0").save()
    d3 = await SoftwareDependency(name="Package3", version="2.5.5").save()

    assert (
        await SoftwareDependency.nodes.order_by(
            RawCypher("toInteger(split($n.version, '.')[0]) DESC"),
        ).all()
    )[0] == d3

    with raises(
        ValueError, match=r"RawCypher: Do not include any action that has side effect"
    ):
        SoftwareDependency.nodes.order_by(
            RawCypher("DETACH DELETE $n"),
        )


@mark_async_test
async def test_extra_filters():
    c1 = await Coffee(name="Icelands finest", price=5, id_=1).save()
    c2 = await Coffee(name="Britains finest", price=10, id_=2).save()
    c3 = await Coffee(name="Japans finest", price=35, id_=3).save()
    c4 = await Coffee(name="US extra-fine", price=None, id_=4).save()

    coffees_5_10 = await Coffee.nodes.filter(price__in=[10, 5])
    assert len(coffees_5_10) == 2, "unexpected number of results"
    assert c1 in coffees_5_10, "doesnt contain 5 price coffee"
    assert c2 in coffees_5_10, "doesnt contain 10 price coffee"

    finest_coffees = await Coffee.nodes.filter(name__iendswith=" Finest")
    assert len(finest_coffees) == 3, "unexpected number of results"
    assert c1 in finest_coffees, "doesnt contain 1st finest coffee"
    assert c2 in finest_coffees, "doesnt contain 2nd finest coffee"
    assert c3 in finest_coffees, "doesnt contain 3rd finest coffee"

    unpriced_coffees = await Coffee.nodes.filter(price__isnull=True)
    assert len(unpriced_coffees) == 1, "unexpected number of results"
    assert c4 in unpriced_coffees, "doesnt contain unpriced coffee"

    coffees_with_id_gte_3 = await Coffee.nodes.filter(id___gte=3)
    assert len(coffees_with_id_gte_3) == 2, "unexpected number of results"
    assert c3 in coffees_with_id_gte_3
    assert c4 in coffees_with_id_gte_3

    with raises(
        ValueError,
        match=r".*Neo4j internals like id or element_id are not allowed for use in this operation.",
    ):
        await Coffee.nodes.filter(elementId="4:xxx:111").all()


def test_traversal_definition_keys_are_valid():
    muckefuck = Coffee(name="Mukkefuck", price=1)

    with raises(ValueError):
        AsyncTraversal(
            muckefuck,
            "a_name",
            {
                "node_class": Supplier,
                "direction": INCOMING,
                "relationship_type": "KNOWS",
                "model": None,
            },
        )

    AsyncTraversal(
        muckefuck,
        "a_name",
        {
            "node_class": Supplier,
            "direction": INCOMING,
            "relation_type": "KNOWS",
            "model": None,
        },
    )


@mark_async_test
async def test_empty_filters():
    """Test this case:
     ```
         SomeModel.nodes.filter().filter(Q(arg1=val1)).all()
         SomeModel.nodes.exclude().exclude(Q(arg1=val1)).all()
         SomeModel.nodes.filter().filter(arg1=val1).all()
    ```
    In django_rest_framework filter uses such as lazy function and
    ``get_queryset`` function in ``GenericAPIView`` should returns
    ``NodeSet`` object.
    """
    c1 = await Coffee(name="Super", price=5, id_=1).save()
    c2 = await Coffee(name="Puper", price=10, id_=2).save()

    empty_filter = Coffee.nodes.filter()

    all_coffees = await empty_filter.all()
    assert len(all_coffees) == 2, "unexpected number of results"

    filter_empty_filter = empty_filter.filter(price=5)
    assert len(await filter_empty_filter.all()) == 1, "unexpected number of results"
    assert (
        c1 in await filter_empty_filter.all()
    ), "doesnt contain c1 in ``filter_empty_filter``"

    filter_q_empty_filter = empty_filter.filter(Q(price=5))
    assert len(await filter_empty_filter.all()) == 1, "unexpected number of results"
    assert (
        c1 in await filter_empty_filter.all()
    ), "doesnt contain c1 in ``filter_empty_filter``"


@mark_async_test
async def test_q_filters():
    c1 = await Coffee(name="Icelands finest", price=5, id_=1).save()
    c2 = await Coffee(name="Britains finest", price=10, id_=2).save()
    c3 = await Coffee(name="Japans finest", price=35, id_=3).save()
    c4 = await Coffee(name="US extra-fine", price=None, id_=4).save()
    c5 = await Coffee(name="Latte", price=35, id_=5).save()
    c6 = await Coffee(name="Cappuccino", price=35, id_=6).save()

    coffees_5_10 = await Coffee.nodes.filter(Q(price=10) | Q(price=5)).all()
    assert len(coffees_5_10) == 2, "unexpected number of results"
    assert c1 in coffees_5_10, "doesnt contain 5 price coffee"
    assert c2 in coffees_5_10, "doesnt contain 10 price coffee"

    coffees_5_6 = (
        await Coffee.nodes.filter(Q(name="Latte") | Q(name="Cappuccino"))
        .filter(price=35)
        .all()
    )
    assert len(coffees_5_6) == 2, "unexpected number of results"
    assert c5 in coffees_5_6, "doesnt contain 5 coffee"
    assert c6 in coffees_5_6, "doesnt contain 6 coffee"

    coffees_5_6 = (
        await Coffee.nodes.filter(price=35)
        .filter(Q(name="Latte") | Q(name="Cappuccino"))
        .all()
    )
    assert len(coffees_5_6) == 2, "unexpected number of results"
    assert c5 in coffees_5_6, "doesnt contain 5 coffee"
    assert c6 in coffees_5_6, "doesnt contain 6 coffee"

    finest_coffees = await Coffee.nodes.filter(name__iendswith=" Finest").all()
    assert len(finest_coffees) == 3, "unexpected number of results"
    assert c1 in finest_coffees, "doesnt contain 1st finest coffee"
    assert c2 in finest_coffees, "doesnt contain 2nd finest coffee"
    assert c3 in finest_coffees, "doesnt contain 3rd finest coffee"

    unpriced_coffees = await Coffee.nodes.filter(Q(price__isnull=True)).all()
    assert len(unpriced_coffees) == 1, "unexpected number of results"
    assert c4 in unpriced_coffees, "doesnt contain unpriced coffee"

    coffees_with_id_gte_3 = await Coffee.nodes.filter(Q(id___gte=3)).all()
    assert len(coffees_with_id_gte_3) == 4, "unexpected number of results"
    assert c3 in coffees_with_id_gte_3
    assert c4 in coffees_with_id_gte_3
    assert c5 in coffees_with_id_gte_3
    assert c6 in coffees_with_id_gte_3

    coffees_5_not_japans = await Coffee.nodes.filter(
        Q(price__gt=5) & ~Q(name="Japans finest")
    ).all()
    assert c3 not in coffees_5_not_japans

    empty_Q_condition = await Coffee.nodes.filter(Q(price=5) | Q()).all()
    assert (
        len(empty_Q_condition) == 1
    ), "undefined Q leading to unexpected number of results"
    assert c1 in empty_Q_condition

    combined_coffees = await Coffee.nodes.filter(
        Q(price=35), Q(name="Latte") | Q(name="Cappuccino")
    ).all()
    assert len(combined_coffees) == 2
    assert c5 in combined_coffees
    assert c6 in combined_coffees
    assert c3 not in combined_coffees

    robusta = await Species(name="Robusta").save()
    await c4.species.connect(robusta)
    latte_or_robusta_coffee = (
        await Coffee.nodes.fetch_relations(Optional("species"))
        .filter(Q(name="Latte") | Q(species__name="Robusta"))
        .all()
    )
    assert len(latte_or_robusta_coffee) == 2

    arabica = await Species(name="Arabica").save()
    await c1.species.connect(arabica)
    robusta_coffee = (
        await Coffee.nodes.fetch_relations(Optional("species"))
        .filter(species__name="Robusta")
        .all()
    )
    # Since the filter is applied on the OPTIONAL MATCH
    # The results will contain all the coffee nodes
    # But only the one connected to the species Robusta will have the species returned
    # Everything else will be None
    assert len(robusta_coffee) == 6
    coffee_with_species = [n[0] for n in robusta_coffee if n[1] is not None]
    assert len(coffee_with_species) == 1
    assert coffee_with_species[0] == c4

    class QQ:
        pass

    with raises(TypeError):
        await Coffee.nodes.filter(Q(price=5) | QQ()).all()


def test_qbase():
    test_print_out = str(Q(price=5) | Q(price=10))
    test_repr = repr(Q(price=5) | Q(price=10))
    assert test_print_out == "(OR: ('price', 5), ('price', 10))"
    assert test_repr == "<Q: (OR: ('price', 5), ('price', 10))>"

    assert ("price", 5) in (Q(price=5) | Q(price=10))

    test_hash = set([Q(price_lt=30) | ~Q(price=5), Q(price_lt=30) | ~Q(price=5)])
    assert len(test_hash) == 1


@mark_async_test
async def test_traversal_filter_left_hand_statement():
    nescafe = await Coffee(name="Nescafe2", price=99).save()
    nescafe_gold = await Coffee(name="Nescafe gold", price=11).save()

    tesco = await Supplier(name="Tesco", delivery_cost=3).save()
    biedronka = await Supplier(name="Biedronka", delivery_cost=5).save()
    lidl = await Supplier(name="Lidl", delivery_cost=3).save()

    await nescafe.suppliers.connect(tesco)
    await nescafe_gold.suppliers.connect(biedronka)
    await nescafe_gold.suppliers.connect(lidl)

    lidl_supplier = (
        await AsyncNodeSet(Coffee.nodes.filter(price=11).suppliers)
        .filter(delivery_cost=3)
        .all()
    )

    assert lidl in lidl_supplier


@mark_async_test
async def test_filter_with_traversal():
    arabica = await Species(name="Arabica").save()
    robusta = await Species(name="Robusta").save()
    nescafe = await Coffee(name="Nescafe", price=11).save()
    nescafe_gold = await Coffee(name="Nescafe Gold", price=99).save()
    tesco = await Supplier(name="Tesco", delivery_cost=3).save()
    await nescafe.suppliers.connect(tesco)
    await nescafe_gold.suppliers.connect(tesco)
    await nescafe.species.connect(arabica)
    await nescafe_gold.species.connect(robusta)

    results = await Coffee.nodes.filter(species__name="Arabica").all()
    assert len(results) == 1
    assert len(results[0]) == 3
    assert results[0][0] == nescafe

    results_multi_hop = await Supplier.nodes.filter(
        coffees__species__name="Arabica"
    ).all()
    assert len(results_multi_hop) == 1
    assert results_multi_hop[0][0] == tesco

    no_results = await Supplier.nodes.filter(coffees__species__name="Noffee").all()
    assert no_results == []


@mark_async_test
async def test_relation_prop_filtering():
    arabica = await Species(name="Arabica").save()
    nescafe = await Coffee(name="Nescafe", price=99).save()
    supplier1 = await Supplier(name="Supplier 1", delivery_cost=3).save()
    supplier2 = await Supplier(name="Supplier 2", delivery_cost=20).save()

    await nescafe.suppliers.connect(supplier1, {"since": datetime(2020, 4, 1, 0, 0)})
    await nescafe.suppliers.connect(supplier2, {"since": datetime(2010, 4, 1, 0, 0)})
    await nescafe.species.connect(arabica)

    result = await Coffee.nodes.filter(
        **{"suppliers|since__gt": datetime(2010, 4, 1, 0, 0)}
    ).all()
    assert len(result) == 1

    results = await Supplier.nodes.filter(
        **{"coffees__name": "Nescafe", "coffees|since__gt": datetime(2018, 4, 1, 0, 0)}
    ).all()

    assert len(results) == 1
    assert results[0][0] == supplier1

    # Test it works with mixed argument syntaxes
    results2 = await Supplier.nodes.filter(
        name="Supplier 1",
        coffees__name="Nescafe",
        **{"coffees|since__gt": datetime(2018, 4, 1, 0, 0)},
    ).all()

    assert len(results2) == 1
    assert results2[0][0] == supplier1


@mark_async_test
async def test_relation_prop_ordering():
    arabica = await Species(name="Arabica").save()
    nescafe = await Coffee(name="Nescafe", price=99).save()
    supplier1 = await Supplier(name="Supplier 1", delivery_cost=3).save()
    supplier2 = await Supplier(name="Supplier 2", delivery_cost=20).save()

    await nescafe.suppliers.connect(supplier1, {"since": datetime(2020, 4, 1, 0, 0)})
    await nescafe.suppliers.connect(supplier2, {"since": datetime(2010, 4, 1, 0, 0)})
    await nescafe.species.connect(arabica)

    results = (
        await Supplier.nodes.fetch_relations("coffees").order_by("-coffees|since").all()
    )
    assert len(results) == 2
    assert results[0][0] == supplier1
    assert results[1][0] == supplier2

    results = (
        await Supplier.nodes.fetch_relations("coffees").order_by("coffees|since").all()
    )
    assert len(results) == 2
    assert results[0][0] == supplier2
    assert results[1][0] == supplier1


@mark_async_test
async def test_fetch_relations():
    arabica = await Species(name="Arabica").save()
    robusta = await Species(name="Robusta").save()
    nescafe = await Coffee(name="Nescafe", price=99).save()
    nescafe_gold = await Coffee(name="Nescafe Gold", price=11).save()

    tesco = await Supplier(name="Tesco", delivery_cost=3).save()
    await nescafe.suppliers.connect(tesco)
    await nescafe_gold.suppliers.connect(tesco)
    await nescafe.species.connect(arabica)

    result = (
        await Supplier.nodes.filter(name="Tesco")
        .fetch_relations("coffees__species")
        .all()
    )
    assert len(result[0]) == 5
    assert arabica in result[0]
    assert robusta not in result[0]
    assert tesco in result[0]
    assert nescafe in result[0]
    assert nescafe_gold not in result[0]

    result = (
        await Species.nodes.filter(name="Robusta")
        .fetch_relations(Optional("coffees__suppliers"))
        .all()
    )
    assert len(result) == 1

    if AsyncUtil.is_async_code:
        count = (
            await Supplier.nodes.filter(name="Tesco")
            .fetch_relations("coffees__species")
            .get_len()
        )
        assert count == 1

        assert (
            await Supplier.nodes.fetch_relations("coffees__species")
            .filter(name="Tesco")
            .check_contains(tesco)
        )
    else:
        count = len(
            Supplier.nodes.filter(name="Tesco")
            .fetch_relations("coffees__species")
            .all()
        )
        assert count == 1

        assert tesco in Supplier.nodes.fetch_relations("coffees__species").filter(
            name="Tesco"
        )


@mark_async_test
async def test_traverse():
    arabica = await Species(name="Arabica").save()
    robusta = await Species(name="Robusta").save()
    nescafe = await Coffee(name="Nescafe", price=99).save()
    nescafe_gold = await Coffee(name="Nescafe Gold", price=11).save()

    tesco = await Supplier(name="Tesco", delivery_cost=3).save()
    await nescafe.suppliers.connect(tesco)
    await nescafe_gold.suppliers.connect(tesco)
    await nescafe.species.connect(arabica)

    result = (
        await Supplier.nodes.filter(name="Tesco")
        .traverse(Path(value="coffees__species", include_rels_in_return=False))
        .all()
    )
    assert len(result[0]) == 3
    assert arabica in result[0]
    assert robusta not in result[0]
    assert tesco in result[0]
    assert nescafe in result[0]
    assert nescafe_gold not in result[0]

    result = (
        await Species.nodes.filter(name="Robusta")
        .traverse(Path(value="coffees__suppliers", optional=True))
        .all()
    )
    assert len(result) == 1

    if AsyncUtil.is_async_code:
        count = (
            await Supplier.nodes.filter(name="Tesco")
            .traverse("coffees__species")
            .get_len()
        )
        assert count == 1

        assert (
            await Supplier.nodes.traverse("coffees__species")
            .filter(name="Tesco")
            .check_contains(tesco)
        )
    else:
        count = len(
            Supplier.nodes.filter(name="Tesco").traverse("coffees__species").all()
        )
        assert count == 1

        assert tesco in Supplier.nodes.traverse("coffees__species").filter(name="Tesco")


@mark_async_test
async def test_traverse_and_order_by():
    arabica = await Species(name="Arabica").save()
    robusta = await Species(name="Robusta").save()
    nescafe = await Coffee(name="Nescafe", price=99).save()
    nescafe_gold = await Coffee(name="Nescafe Gold", price=110).save()
    tesco = await Supplier(name="Tesco", delivery_cost=3).save()
    await nescafe.suppliers.connect(tesco)
    await nescafe_gold.suppliers.connect(tesco)
    await nescafe.species.connect(arabica)
    await nescafe_gold.species.connect(robusta)

    results = (
        await Species.nodes.fetch_relations("coffees").order_by("-coffees__price").all()
    )
    assert len(results) == 2
    assert len(results[0]) == 3  # 2 nodes and 1 relation
    assert results[0][0] == robusta
    assert results[1][0] == arabica


@mark_async_test
async def test_annotate_and_collect():
    arabica = await Species(name="Arabica").save()
    robusta = await Species(name="Robusta").save()
    nescafe = await Coffee(name="Nescafe 1002", price=99).save()
    nescafe_gold = await Coffee(name="Nescafe 1003", price=11).save()

    tesco = await Supplier(name="Tesco", delivery_cost=3).save()
    await nescafe.suppliers.connect(tesco)
    await nescafe_gold.suppliers.connect(tesco)
    await nescafe.species.connect(arabica)
    await nescafe_gold.species.connect(robusta)
    await nescafe_gold.species.connect(arabica)

    result = (
        await Supplier.nodes.traverse_relations(species="coffees__species")
        .annotate(Collect("species"))
        .all()
    )
    assert len(result) == 1
    assert len(result[0][1][0]) == 3  # 3 species must be there (with 2 duplicates)

    result = (
        await Supplier.nodes.traverse_relations(species="coffees__species")
        .annotate(Collect("species", distinct=True))
        .all()
    )
    assert len(result[0][1][0]) == 2  # 2 species must be there

    result = (
        await Supplier.nodes.traverse_relations(species="coffees__species")
        .annotate(Size(Collect("species", distinct=True)))
        .all()
    )
    assert result[0][1] == 2  # 2 species

    result = (
        await Supplier.nodes.traverse_relations(species="coffees__species")
        .annotate(all_species=Collect("species", distinct=True))
        .all()
    )
    assert len(result[0][1][0]) == 2  # 2 species must be there

    result = (
        await Supplier.nodes.traverse_relations("coffees__species")
        .annotate(
            all_species=Collect(NodeNameResolver("coffees__species"), distinct=True),
            all_species_rels=Collect(
                RelationNameResolver("coffees__species"), distinct=True
            ),
        )
        .all()
    )
    assert len(result[0][1][0]) == 2  # 2 species must be there
    assert len(result[0][2][0]) == 3  # 3 species relations must be there


@mark_async_test
async def test_resolve_subgraph():
    arabica = await Species(name="Arabica").save()
    robusta = await Species(name="Robusta").save()
    nescafe = await Coffee(name="Nescafe", price=99).save()
    nescafe_gold = await Coffee(name="Nescafe Gold", price=11).save()

    tesco = await Supplier(name="Tesco", delivery_cost=3).save()
    await nescafe.suppliers.connect(tesco)
    await nescafe_gold.suppliers.connect(tesco)
    await nescafe.species.connect(arabica)
    await nescafe_gold.species.connect(robusta)

    with raises(
        RuntimeError,
        match=re.escape(
            "Nothing to resolve. Make sure to include relations in the result using fetch_relations() or filter()."
        ),
    ):
        result = await Supplier.nodes.resolve_subgraph()

    with raises(
        NotImplementedError,
        match=re.escape(
            "You cannot use traverse_relations() with resolve_subgraph(), use fetch_relations() instead."
        ),
    ):
        result = await Supplier.nodes.traverse_relations(
            "coffees__species"
        ).resolve_subgraph()

    result = await Supplier.nodes.fetch_relations("coffees__species").resolve_subgraph()
    assert len(result) == 2

    assert hasattr(result[0], "_relations")
    assert "coffees" in result[0]._relations
    coffees = result[0]._relations["coffees"]
    assert hasattr(coffees, "_relations")
    assert "species" in coffees._relations

    assert hasattr(result[1], "_relations")
    assert "coffees" in result[1]._relations
    coffees = result[1]._relations["coffees"]
    assert hasattr(coffees, "_relations")
    assert "species" in coffees._relations


@mark_async_test
async def test_resolve_subgraph_optional():
    arabica = await Species(name="Arabica").save()
    nescafe = await Coffee(name="Nescafe", price=99).save()
    nescafe_gold = await Coffee(name="Nescafe Gold", price=11).save()

    tesco = await Supplier(name="Tesco", delivery_cost=3).save()
    await nescafe.suppliers.connect(tesco)
    await nescafe_gold.suppliers.connect(tesco)
    await nescafe.species.connect(arabica)

    result = await Supplier.nodes.fetch_relations(
        Optional("coffees__species")
    ).resolve_subgraph()
    assert len(result) == 1

    assert hasattr(result[0], "_relations")
    assert "coffees" in result[0]._relations
    coffees = result[0]._relations["coffees"]
    assert hasattr(coffees, "_relations")
    assert "species" in coffees._relations
    assert coffees._relations["species"] == arabica


@mark_async_test
async def test_subquery():
    arabica = await Species(name="Arabica").save()
    nescafe = await Coffee(name="Nescafe", price=99).save()
    supplier1 = await Supplier(name="Supplier 1", delivery_cost=3).save()
    supplier2 = await Supplier(name="Supplier 2", delivery_cost=20).save()

    await nescafe.suppliers.connect(supplier1)
    await nescafe.suppliers.connect(supplier2)
    await nescafe.species.connect(arabica)

    subquery = await Coffee.nodes.subquery(
        Coffee.nodes.traverse_relations(suppliers="suppliers")
        .intermediate_transform(
            {"suppliers": {"source": "suppliers"}}, ordering=["suppliers.delivery_cost"]
        )
        .annotate(supps=Last(Collect("suppliers"))),
        ["supps"],
        [NodeNameResolver("self")],
    )
    result = await subquery.all()
    assert len(result) == 1
    assert len(result[0]) == 2
    assert result[0][0] == supplier2

    with raises(
        RuntimeError,
        match=re.escape("Variable 'unknown' is not returned by subquery."),
    ):
        result = await Coffee.nodes.subquery(
            Coffee.nodes.traverse_relations(suppliers="suppliers").annotate(
                supps=Collect("suppliers")
            ),
            ["unknown"],
        )

    result_string_context = await subquery.subquery(
        Coffee.nodes.traverse_relations(supps2="suppliers").annotate(
            supps2=Collect("supps")
        ),
        ["supps2"],
        ["supps"],
    )
    result_string_context = await result_string_context.all()
    assert len(result) == 1
    additional_elements = [
        item for item in result_string_context[0] if item not in result[0]
    ]
    assert len(additional_elements) == 1
    assert isinstance(additional_elements[0], list)

    with raises(ValueError, match=r"Wrong variable specified in initial context"):
        result = await Coffee.nodes.subquery(
            Coffee.nodes.traverse_relations(suppliers="suppliers").annotate(
                supps=Collect("suppliers")
            ),
            ["supps"],
            [2],
        )


@mark_async_test
async def test_subquery_other_node():
    arabica = await Species(name="Arabica").save()
    nescafe = await Coffee(name="Nescafe", price=99).save()
    supplier1 = await Supplier(name="Supplier 1", delivery_cost=3).save()
    supplier2 = await Supplier(name="Supplier 2", delivery_cost=20).save()

    await nescafe.suppliers.connect(supplier1)
    await nescafe.suppliers.connect(supplier2)
    await nescafe.species.connect(arabica)

    result = await Coffee.nodes.subquery(
        Supplier.nodes.filter(name="Supplier 2").intermediate_transform(
            {
                "cost": {
                    "source": "supplier",
                    "source_prop": "delivery_cost",
                    "include_in_return": True,
                }
            }
        ),
        ["cost"],
    )
    result = await result.all()
    assert len(result) == 1
    assert result[0][0] == 20


@mark_async_test
async def test_intermediate_transform():
    arabica = await Species(name="Arabica").save()
    nescafe = await Coffee(name="Nescafe", price=99).save()
    supplier1 = await Supplier(name="Supplier 1", delivery_cost=3).save()
    supplier2 = await Supplier(name="Supplier 2", delivery_cost=20).save()

    await nescafe.suppliers.connect(supplier1, {"since": datetime(2020, 4, 1, 0, 0)})
    await nescafe.suppliers.connect(supplier2, {"since": datetime(2010, 4, 1, 0, 0)})
    await nescafe.species.connect(arabica)

    result = (
        await Coffee.nodes.fetch_relations("suppliers")
        .intermediate_transform(
            {
                "coffee": {"source": "coffee", "include_in_return": True},
                "suppliers": {"source": NodeNameResolver("suppliers")},
                "r": {"source": RelationNameResolver("suppliers")},
                "cost": {
                    "source": NodeNameResolver("suppliers"),
                    "source_prop": "delivery_cost",
                },
            },
            distinct=True,
            ordering=["-r.since"],
        )
        .annotate(oldest_supplier=Last(Collect("suppliers")))
        .all()
    )

    assert len(result) == 1
    assert result[0][0] == nescafe
    assert result[0][1] == supplier2

    with raises(
        ValueError,
        match=re.escape(
            r"Wrong source type specified for variable 'test', should be a string or an instance of NodeNameResolver or RelationNameResolver"
        ),
    ):
        Coffee.nodes.traverse_relations(suppliers="suppliers").intermediate_transform(
            {
                "test": {"source": Collect("suppliers")},
            }
        )
    with raises(
        ValueError,
        match=re.escape(
            r"You must provide one variable at least when calling intermediate_transform()"
        ),
    ):
        Coffee.nodes.traverse_relations(suppliers="suppliers").intermediate_transform(
            {}
        )


@mark_async_test
async def test_mix_functions():
    # Test with a mix of all advanced querying functions

    eiffel_tower = await Building(name="Eiffel Tower").save()
    empire_state_building = await Building(name="Empire State Building").save()
    miranda = await Student(name="Miranda").save()
    await miranda.lives_in.connect(empire_state_building)
    jean_pierre = await Student(name="Jean-Pierre").save()
    await jean_pierre.lives_in.connect(eiffel_tower)
    mireille = await Student(name="Mireille").save()
    mimoun_jr = await Student(name="Mimoun Jr").save()
    mimoun = await Student(name="Mimoun").save()
    await mireille.lives_in.connect(eiffel_tower)
    await mimoun_jr.lives_in.connect(eiffel_tower)
    await mimoun.lives_in.connect(eiffel_tower)
    await mimoun.parents.connect(mireille)
    await mimoun.children.connect(mimoun_jr)
    math = await Course(name="Math").save()
    dessin = await Course(name="Dessin").save()
    await mimoun.courses.connect(
        math,
        {
            "level": "1.2",
            "start_date": datetime(2020, 6, 2),
            "end_date": datetime(2020, 12, 31),
        },
    )
    await mimoun.courses.connect(
        math,
        {
            "level": "1.1",
            "start_date": datetime(2020, 1, 1),
            "end_date": datetime(2020, 6, 1),
        },
    )
    await mimoun_jr.courses.connect(
        math,
        {
            "level": "1.1",
            "start_date": datetime(2020, 1, 1),
            "end_date": datetime(2020, 6, 1),
        },
    )

    await mimoun_jr.preferred_course.connect(dessin)

    full_nodeset = (
        await Student.nodes.filter(name__istartswith="m", lives_in__name="Eiffel Tower")
        .order_by("name")
        .fetch_relations(
            "parents",
            Optional("children__preferred_course"),
        )
        .subquery(
            Student.nodes.fetch_relations("courses")
            .intermediate_transform(
                {"rel": {"source": RelationNameResolver("courses")}},
                ordering=[
                    RawCypher("toInteger(split(rel.level, '.')[0])"),
                    RawCypher("toInteger(split(rel.level, '.')[1])"),
                    "rel.end_date",
                    "rel.start_date",
                ],
            )
            .annotate(
                latest_course=Last(Collect("rel")),
            ),
            ["latest_course"],
        )
    )

    subgraph = await full_nodeset.annotate(
        children=Collect(NodeNameResolver("children"), distinct=True),
        children_preferred_course=Collect(
            NodeNameResolver("children__preferred_course"), distinct=True
        ),
    ).resolve_subgraph()

    assert len(subgraph) == 2
    assert subgraph[0] == mimoun
    assert subgraph[1] == mimoun_jr
    mimoun_returned_rels = subgraph[0]._relations
    assert mimoun_returned_rels["children"] == mimoun_jr
    assert mimoun_returned_rels["children"]._relations["preferred_course"] == dessin
    assert mimoun_returned_rels["parents"] == mireille
    assert mimoun_returned_rels["latest_course_relationship"].level == "1.2"


@mark_async_test
async def test_issue_795():
    jim = await PersonX(name="Jim", age=3).save()  # Create
    jim.age = 4
    await jim.save()  # Update, (with validation)

    germany = await CountryX(code="DE").save()
    await jim.country.connect(germany)
    berlin = await CityX(name="Berlin").save()
    await berlin.country.connect(germany)
    await jim.city.connect(berlin)

    with raises(
        RelationshipClassNotDefined,
        match=r"[\s\S]*Note that when using the fetch_relations method, the relationship type must be defined in the model.*",
    ):
        _ = await PersonX.nodes.fetch_relations("country").all()


@mark_async_test
async def test_in_filter_with_array_property():
    tags = ["smoother", "sweeter", "chocolate", "sugar"]
    no_match = ["organic"]
    arabica = await Species(name="Arabica", tags=tags).save()

    assert arabica in await Species.nodes.filter(
        tags__in=tags
    ), "Species not found by tags given"
    assert arabica in await Species.nodes.filter(
        Q(tags__in=tags)
    ), "Species not found with Q by tags given"
    assert arabica not in await Species.nodes.filter(
        ~Q(tags__in=tags)
    ), "Species found by tags given in negated query"
    assert arabica not in await Species.nodes.filter(
        tags__in=no_match
    ), "Species found by tags with not match tags given"


@mark_async_test
async def test_unique_variables():
    arabica = await Species(name="Arabica").save()
    nescafe = await Coffee(name="Nescafe", price=99).save()
    gold3000 = await Coffee(name="Gold 3000", price=11).save()
    supplier1 = await Supplier(name="Supplier 1", delivery_cost=3).save()
    supplier2 = await Supplier(name="Supplier 2", delivery_cost=20).save()
    supplier3 = await Supplier(name="Supplier 3", delivery_cost=20).save()

    await nescafe.suppliers.connect(supplier1, {"since": datetime(2020, 4, 1, 0, 0)})
    await nescafe.suppliers.connect(supplier2, {"since": datetime(2010, 4, 1, 0, 0)})
    await nescafe.species.connect(arabica)
    await gold3000.suppliers.connect(supplier1, {"since": datetime(2020, 4, 1, 0, 0)})
    await gold3000.species.connect(arabica)

    nodeset = Supplier.nodes.fetch_relations("coffees", "coffees__species").filter(
        coffees__name="Nescafe"
    )
    ast = await nodeset.query_cls(nodeset).build_ast()
    query = ast.build_query()
    assert "coffee_coffees1" in query
    assert "coffee_coffees2" in query
    results = await nodeset.all()
    # This will be 3 because 2 suppliers for Nescafe and 1 for Gold 3000
    # Gold 3000 is traversed because coffees__species redefines the coffees traversal
    assert len(results) == 3

    nodeset = (
        Supplier.nodes.fetch_relations("coffees", "coffees__species")
        .filter(coffees__name="Nescafe")
        .unique_variables("coffees")
    )
    ast = await nodeset.query_cls(nodeset).build_ast()
    query = ast.build_query()
    assert "coffee_coffees" in query
    assert "coffee_coffees1" not in query
    assert "coffee_coffees2" not in query
    results = await nodeset.all()
    # This will 2 because Gold 3000 is excluded this time
    # As coffees will be reused in coffees__species
    assert len(results) == 2


@mark_async_test
async def test_async_iterator():
    n = 10
    if AsyncUtil.is_async_code:
        for i in range(n):
            await Coffee(name=f"xxx_{i}", price=i).save()

        nodes = await Coffee.nodes
        # assert that nodes was created
        assert isinstance(nodes, list)
        assert all(isinstance(i, Coffee) for i in nodes)
        assert len(nodes) == n

        counter = 0
        async for node in Coffee.nodes:
            assert isinstance(node, Coffee)
            counter += 1

        # assert that generator runs loop above
        assert counter == n

        counter = 0
        for node in await Coffee.nodes:
            assert isinstance(node, Coffee)
            counter += 1

        # assert that generator runs loop above
        assert counter == n


def assert_last_query_startswith(mock_func, query) -> bool:
    return mock_func.call_args_list[-1].kwargs["query"].startswith(query)


@mark_async_test
async def test_parallel_runtime(mocker):
    if (
        not await adb.version_is_higher_than("5.13")
        or not await adb.edition_is_enterprise()
    ):
        skip("Only supported for Enterprise 5.13 and above.")

    assert await adb.parallel_runtime_available()

    # Parallel should be applied to custom Cypher query
    async with adb.parallel_read_transaction:
        # Mock transaction.run to access executed query
        # Assert query starts with CYPHER runtime=parallel
        assert adb._parallel_runtime == True
        mock_transaction_run = mocker.patch("neo4j.AsyncTransaction.run")
        await adb.cypher_query("MATCH (n:Coffee) RETURN n")
        assert assert_last_query_startswith(
            mock_transaction_run, "CYPHER runtime=parallel"
        )
    # Test exiting the context sets the parallel_runtime to False
    assert adb._parallel_runtime == False

    # Parallel should be applied to neomodel queries
    async with adb.parallel_read_transaction:
        mock_transaction_run_2 = mocker.patch("neo4j.AsyncTransaction.run")
        await Coffee.nodes.all()
        assert assert_last_query_startswith(
            mock_transaction_run_2, "CYPHER runtime=parallel"
        )


@mark_async_test
async def test_parallel_runtime_conflict(mocker):
    if await adb.version_is_higher_than("5.13") and await adb.edition_is_enterprise():
        skip("Test for unavailable parallel runtime.")

    assert not await adb.parallel_runtime_available()
    mock_transaction_run = mocker.patch("neo4j.AsyncTransaction.run")
    with warns(
        UserWarning,
        match="Parallel runtime is only available in Neo4j Enterprise Edition 5.13",
    ):
        async with adb.parallel_read_transaction:
            await Coffee.nodes.all()
            assert not adb._parallel_runtime
            assert not assert_last_query_startswith(
                mock_transaction_run, "CYPHER runtime=parallel"
            )



from test._async_compat import mark_async_test

from pytest import raises, skip

from neomodel import (
    AsyncRelationshipTo,
    AsyncStructuredNode,
    AsyncStructuredRel,
    DateProperty,
    IntegerProperty,
    StringProperty,
    adb,
    config,
)
from neomodel.exceptions import (
    NodeClassAlreadyDefined,
    NodeClassNotDefined,
    RelationshipClassRedefined,
)


@mark_async_test
async def test_db_specific_node_labels():
    if not await adb.edition_is_enterprise():
        skip("Skipping test for community edition")
    db_one = "one"
    db_two = "two"
    await adb.cypher_query(f"CREATE DATABASE {db_one} IF NOT EXISTS")
    await adb.cypher_query(f"CREATE DATABASE {db_two} IF NOT EXISTS")

    class Experiment(AsyncStructuredNode):
        __label__ = "Experiment"
        name = StringProperty()

    class PatientOne(AsyncStructuredNode):
        __label__ = "Patient"
        __target_databases__ = [db_one]
        name = StringProperty()

    class PatientTwo(AsyncStructuredNode):
        __label__ = "Patient"
        __target_databases__ = [db_two]
        identifier = StringProperty()

    # This should have reached this point without failing
    # It means db specific registry is allowing reuse of labels in different databases
    # Next test will check if the standard registry still denies reuse of labels
    with raises(NodeClassAlreadyDefined):

        class ExperimentTwo(AsyncStructuredNode):
            __label__ = "Experiment"
            name = StringProperty()

        await ExperimentTwo(name="experiment2").save()

    # Finally, this tests that db specific registry denies reuse of labels in the same db
    with raises(NodeClassAlreadyDefined):

        class PatientOneBis(AsyncStructuredNode):
            __label__ = "Patient"
            __target_databases__ = [db_one]
            name = StringProperty()

        await PatientOneBis(name="patient1.2").save()

    # Now, we will test object resolution
    await adb.close_connection()
    await adb.set_connection(url=f"{config.DATABASE_URL}/{db_one}")
    await adb.clear_neo4j_database()
    patient1 = await PatientOne(name="patient1").save()
    patients, _ = await adb.cypher_query(
        "MATCH (n:Patient) RETURN n", resolve_objects=True
    )
    # This means that the auto object resolution is working
    assert patients[0][0] == patient1

    await adb.close_connection()
    await adb.set_connection(url=f"{config.DATABASE_URL}/{db_two}")
    await adb.clear_neo4j_database()
    patient2 = await PatientTwo(identifier="patient2").save()
    patients, _ = await adb.cypher_query(
        "MATCH (n:Patient) RETURN n", resolve_objects=True
    )
    assert patients[0][0] == patient2

    await adb.close_connection()
    await adb.set_connection(url=config.DATABASE_URL)


@mark_async_test
async def test_resolution_not_defined_class():
    if not await adb.edition_is_enterprise():
        skip("Skipping test for community edition")

    class PatientX(AsyncStructuredNode):
        __label__ = "Patient"
        __target_databases__ = ["db_x"]
        name = StringProperty()

    await adb.cypher_query("CREATE (n:Gabagool)")
    with raises(
        NodeClassNotDefined,
        match=r"Node with labels Gabagool does not resolve to any of the known objects[\s\S]*Database-specific: db_x.*",
    ):
        _ = await adb.cypher_query("MATCH (n:Gabagool) RETURN n", resolve_objects=True)



"""
Provides integration with `numpy <https://numpy.org/>`_.

.. note::
   This module requires numpy to be installed, and will raise a
   warning if this is not available.

Example:

    >>> from neomodel.async_ import db
    >>> from neomodel.integration.numpy import to_nparray
    >>> db.set_connection('bolt://neo4j:secret@localhost:7687')
    >>> df = to_nparray(db.cypher_query("MATCH (u:User) RETURN u.email AS email, u.name AS name"))
    >>> df
    array([['jimla@test.com', 'jimla'], ['jimlo@test.com', 'jimlo']])
"""


from warnings import warn

try:
    # noinspection PyPackageRequirements
    from numpy import array as nparray
except ImportError:
    warn(
        "The neomodel.integration.numpy module expects numpy to be installed "
        "but it does not appear to be available."
    )
    raise


def to_ndarray(query_results: tuple, dtype=None, order="K"):
    """Convert the results of a db.cypher_query call into a numpy array.
    Optionally, specify a datatype and/or an order for the columns.
    """
    results, _ = query_results
    return nparray(results, dtype=dtype, order=order)



from test._async_compat import mark_sync_test

import pytest
from neo4j.exceptions import AuthError

from neomodel import (
    IntegerProperty,
    RelationshipTo,
    StringProperty,
    StructuredNode,
    StructuredRel,
    db,
)


class City(StructuredNode):
    name = StringProperty()


class InCity(StructuredRel):
    creation_year = IntegerProperty(index=True)


class Venue(StructuredNode):
    name = StringProperty(unique_index=True)
    creator = StringProperty(index=True)
    in_city = RelationshipTo(City, relation_type="IN", model=InCity)


@mark_sync_test
def test_clear_database():
    venue = Venue(name="Royal Albert Hall", creator="Queen Victoria").save()
    city = City(name="London").save()
    venue.in_city.connect(city)

    # Clear only the data
    db.clear_neo4j_database()
    database_is_populated, _ = db.cypher_query(
        "MATCH (a) return count(a)>0 as database_is_populated"
    )

    assert database_is_populated[0][0] is False

    db.install_all_labels()
    indexes = db.list_indexes(exclude_token_lookup=True)
    constraints = db.list_constraints()
    assert len(indexes) > 0
    assert len(constraints) > 0

    # Clear constraints and indexes too
    db.clear_neo4j_database(clear_constraints=True, clear_indexes=True)

    indexes = db.list_indexes(exclude_token_lookup=True)
    constraints = db.list_constraints()
    assert len(indexes) == 0
    assert len(constraints) == 0


@mark_sync_test
def test_change_password():
    prev_password = "foobarbaz"
    new_password = "newpassword"
    prev_url = f"bolt://neo4j:{prev_password}@localhost:7687"
    new_url = f"bolt://neo4j:{new_password}@localhost:7687"

    db.change_neo4j_password("neo4j", new_password)
    db.close_connection()

    db.set_connection(url=new_url)
    db.close_connection()

    with pytest.raises(AuthError):
        db.set_connection(url=prev_url)

    db.close_connection()

    db.set_connection(url=new_url)
    db.change_neo4j_password("neo4j", prev_password)
    db.close_connection()

    db.set_connection(url=prev_url)



from .mark_decorator import (
    AsyncTestDecorators,
    TestDecorators,
    mark_async_function_auto_fixture,
    mark_async_session_auto_fixture,
    mark_async_test,
    mark_sync_function_auto_fixture,
    mark_sync_session_auto_fixture,
    mark_sync_test,
)

__all__ = [
    "AsyncTestDecorators",
    "mark_async_test",
    "mark_sync_test",
    "TestDecorators",
    "mark_async_session_auto_fixture",
    "mark_async_function_auto_fixture",
    "mark_sync_session_auto_fixture",
    "mark_sync_function_auto_fixture",
]



from datetime import datetime
from test._async_compat import mark_async_test

from pytest import raises

from neomodel import (
    AsyncStructuredNode,
    AsyncStructuredRel,
    DateProperty,
    IntegerProperty,
    StringProperty,
    adb,
)
from neomodel.exceptions import RequiredProperty, UniqueProperty


class User(AsyncStructuredNode):
    email = StringProperty(unique_index=True, required=True)
    age = IntegerProperty(index=True)

    @property
    def email_alias(self):
        return self.email

    @email_alias.setter  # noqa
    def email_alias(self, value):
        self.email = value


class NodeWithoutProperty(AsyncStructuredNode):
    pass


@mark_async_test
async def test_issue_233():
    class BaseIssue233(AsyncStructuredNode):
        __abstract_node__ = True

        def __getitem__(self, item):
            return self.__dict__[item]

    class Issue233(BaseIssue233):
        uid = StringProperty(unique_index=True, required=True)

    i = await Issue233(uid="testgetitem").save()
    assert i["uid"] == "testgetitem"


def test_issue_72():
    user = User(email="foo@bar.com")
    assert user.age is None


@mark_async_test
async def test_required():
    with raises(RequiredProperty):
        await User(age=3).save()


def test_repr_and_str():
    u = User(email="robin@test.com", age=3)
    assert repr(u) == "<User: {'email': 'robin@test.com', 'age': 3}>"
    assert str(u) == "{'email': 'robin@test.com', 'age': 3}"


@mark_async_test
async def test_get_and_get_or_none():
    u = User(email="robin@test.com", age=3)
    assert await u.save()
    rob = await User.nodes.get(email="robin@test.com")
    assert rob.email == "robin@test.com"
    assert rob.age == 3

    rob = await User.nodes.get_or_none(email="robin@test.com")
    assert rob.email == "robin@test.com"

    n = await User.nodes.get_or_none(email="robin@nothere.com")
    assert n is None


@mark_async_test
async def test_first_and_first_or_none():
    u = User(email="matt@test.com", age=24)
    assert await u.save()
    u2 = User(email="tbrady@test.com", age=40)
    assert await u2.save()
    tbrady = await User.nodes.order_by("-age").first()
    assert tbrady.email == "tbrady@test.com"
    assert tbrady.age == 40

    tbrady = await User.nodes.order_by("-age").first_or_none()
    assert tbrady.email == "tbrady@test.com"

    n = await User.nodes.first_or_none(email="matt@nothere.com")
    assert n is None


def test_bare_init_without_save():
    """
    If a node model is initialised without being saved, accessing its `element_id` should
    return None.
    """
    assert User().element_id is None


@mark_async_test
async def test_save_to_model():
    u = User(email="jim@test.com", age=3)
    assert await u.save()
    assert u.element_id is not None
    assert u.email == "jim@test.com"
    assert u.age == 3


@mark_async_test
async def test_save_node_without_properties():
    n = NodeWithoutProperty()
    assert await n.save()
    assert n.element_id is not None


@mark_async_test
async def test_unique():
    await adb.install_labels(User)
    await User(email="jim1@test.com", age=3).save()
    with raises(UniqueProperty):
        await User(email="jim1@test.com", age=3).save()


@mark_async_test
async def test_update_unique():
    u = await User(email="jimxx@test.com", age=3).save()
    await u.save()  # this shouldn't fail


@mark_async_test
async def test_update():
    user = await User(email="jim2@test.com", age=3).save()
    assert user
    user.email = "jim2000@test.com"
    await user.save()
    jim = await User.nodes.get(email="jim2000@test.com")
    assert jim
    assert jim.email == "jim2000@test.com"


@mark_async_test
async def test_save_through_magic_property():
    user = await User(email_alias="blah@test.com", age=8).save()
    assert user.email_alias == "blah@test.com"
    user = await User.nodes.get(email="blah@test.com")
    assert user.email == "blah@test.com"
    assert user.email_alias == "blah@test.com"

    user1 = await User(email="blah1@test.com", age=8).save()
    assert user1.email_alias == "blah1@test.com"
    user1.email_alias = "blah2@test.com"
    assert await user1.save()
    user2 = await User.nodes.get(email="blah2@test.com")
    assert user2


class Customer2(AsyncStructuredNode):
    __label__ = "customers"
    email = StringProperty(unique_index=True, required=True)
    age = IntegerProperty(index=True)


@mark_async_test
async def test_not_updated_on_unique_error():
    await adb.install_labels(Customer2)
    await Customer2(email="jim@bob.com", age=7).save()
    test = await Customer2(email="jim1@bob.com", age=2).save()
    test.email = "jim@bob.com"
    with raises(UniqueProperty):
        await test.save()
    customers = await Customer2.nodes
    assert customers[0].email != customers[1].email
    assert (await Customer2.nodes.get(email="jim@bob.com")).age == 7
    assert (await Customer2.nodes.get(email="jim1@bob.com")).age == 2


@mark_async_test
async def test_label_not_inherited():
    class Customer3(Customer2):
        address = StringProperty()

    assert Customer3.__label__ == "Customer3"
    c = await Customer3(email="test@test.com").save()
    assert "customers" in await c.labels()
    assert "Customer3" in await c.labels()

    c = await Customer2.nodes.get(email="test@test.com")
    assert isinstance(c, Customer2)
    assert "customers" in await c.labels()
    assert "Customer3" in await c.labels()


@mark_async_test
async def test_refresh():
    c = await Customer2(email="my@email.com", age=16).save()
    c.my_custom_prop = "value"
    copy = await Customer2.nodes.get(email="my@email.com")
    copy.age = 20
    await copy.save()

    assert c.age == 16

    await c.refresh()
    assert c.age == 20
    assert c.my_custom_prop == "value"

    c = Customer2.inflate(c.element_id)
    c.age = 30
    await c.refresh()

    assert c.age == 20

    _db_version = await adb.database_version
    if _db_version.startswith("4"):
        c = Customer2.inflate(999)
    else:
        c = Customer2.inflate("4:xxxxxx:999")
    with raises(Customer2.DoesNotExist):
        await c.refresh()


@mark_async_test
async def test_setting_value_to_none():
    c = await Customer2(email="alice@bob.com", age=42).save()
    assert c.age is not None

    c.age = None
    await c.save()

    copy = await Customer2.nodes.get(email="alice@bob.com")
    assert copy.age is None


@mark_async_test
async def test_inheritance():
    class User(AsyncStructuredNode):
        __abstract_node__ = True
        name = StringProperty(unique_index=True)

    class Shopper(User):
        balance = IntegerProperty(index=True)

        async def credit_account(self, amount):
            self.balance = self.balance + int(amount)
            await self.save()

    jim = await Shopper(name="jimmy", balance=300).save()
    await jim.credit_account(50)

    assert Shopper.__label__ == "Shopper"
    assert jim.balance == 350
    assert len(jim.inherited_labels()) == 1
    assert len(await jim.labels()) == 1
    assert (await jim.labels())[0] == "Shopper"


@mark_async_test
async def test_inherited_optional_labels():
    class BaseOptional(AsyncStructuredNode):
        __optional_labels__ = ["Alive"]
        name = StringProperty(unique_index=True)

    class ExtendedOptional(BaseOptional):
        __optional_labels__ = ["RewardsMember"]
        balance = IntegerProperty(index=True)

        async def credit_account(self, amount):
            self.balance = self.balance + int(amount)
            await self.save()

    henry = await ExtendedOptional(name="henry", balance=300).save()
    await henry.credit_account(50)

    assert ExtendedOptional.__label__ == "ExtendedOptional"
    assert henry.balance == 350
    assert len(henry.inherited_labels()) == 2
    assert len(await henry.labels()) == 2

    assert set(henry.inherited_optional_labels()) == {"Alive", "RewardsMember"}


@mark_async_test
async def test_mixins():
    class UserMixin:
        name = StringProperty(unique_index=True)
        password = StringProperty()

    class CreditMixin:
        balance = IntegerProperty(index=True)

        async def credit_account(self, amount):
            self.balance = self.balance + int(amount)
            await self.save()

    class Shopper2(AsyncStructuredNode, UserMixin, CreditMixin):
        pass

    jim = await Shopper2(name="jimmy", balance=300).save()
    await jim.credit_account(50)

    assert Shopper2.__label__ == "Shopper2"
    assert jim.balance == 350
    assert len(jim.inherited_labels()) == 1
    assert len(await jim.labels()) == 1
    assert (await jim.labels())[0] == "Shopper2"


@mark_async_test
async def test_date_property():
    class DateTest(AsyncStructuredNode):
        birthdate = DateProperty()

    user = await DateTest(birthdate=datetime.now()).save()


def test_reserved_property_keys():
    error_match = r".*is not allowed as it conflicts with neomodel internals.*"
    with raises(ValueError, match=error_match):

        class ReservedPropertiesDeletedNode(AsyncStructuredNode):
            deleted = StringProperty()

    with raises(ValueError, match=error_match):

        class ReservedPropertiesIdNode(AsyncStructuredNode):
            id = StringProperty()

    with raises(ValueError, match=error_match):

        class ReservedPropertiesElementIdNode(AsyncStructuredNode):
            element_id = StringProperty()

    with raises(ValueError, match=error_match):

        class ReservedPropertiesIdRel(AsyncStructuredRel):
            id = StringProperty()

    with raises(ValueError, match=error_match):

        class ReservedPropertiesElementIdRel(AsyncStructuredRel):
            element_id = StringProperty()

    error_match = r"Property names 'source' and 'target' are not allowed as they conflict with neomodel internals."
    with raises(ValueError, match=error_match):

        class ReservedPropertiesSourceRel(AsyncStructuredRel):
            source = StringProperty()

    with raises(ValueError, match=error_match):

        class ReservedPropertiesTargetRel(AsyncStructuredRel):
            target = StringProperty()



from test._async_compat import mark_async_test

import pytest
from neo4j.exceptions import ClientError
from pytest import raises

from neomodel import adb
from neomodel.exceptions import FeatureNotSupported


@mark_async_test
async def test_impersonate():
    if not await adb.edition_is_enterprise():
        pytest.skip("Skipping test for community edition")
    with await adb.impersonate(user="troygreene"):
        results, _ = await adb.cypher_query("RETURN 'Doo Wacko !'")
        assert results[0][0] == "Doo Wacko !"


@mark_async_test
async def test_impersonate_unauthorized():
    if not await adb.edition_is_enterprise():
        pytest.skip("Skipping test for community edition")
    with await adb.impersonate(user="unknownuser"):
        with raises(ClientError):
            _ = await adb.cypher_query("RETURN 'Gabagool'")


@mark_async_test
async def test_impersonate_multiple_transactions():
    if not await adb.edition_is_enterprise():
        pytest.skip("Skipping test for community edition")
    with await adb.impersonate(user="troygreene"):
        async with adb.transaction:
            results, _ = await adb.cypher_query("RETURN 'Doo Wacko !'")
            assert results[0][0] == "Doo Wacko !"

        async with adb.transaction:
            results, _ = await adb.cypher_query("SHOW CURRENT USER")
            assert results[0][0] == "troygreene"

    results, _ = await adb.cypher_query("SHOW CURRENT USER")
    assert results[0][0] == "neo4j"


@mark_async_test
async def test_impersonate_community():
    if await adb.edition_is_enterprise():
        pytest.skip("Skipping test for enterprise edition")
    with raises(FeatureNotSupported):
        with await adb.impersonate(user="troygreene"):
            _ = await adb.cypher_query("RETURN 'Gabagoogoo'")



from typing import List

class VectorFilter(object):
    """
    Represents a CALL db.index.vector.query* neo functions call within the OGM
    
    :param topk: The amount of objects returned by this vector filter. 
    :type topk: int
    :param vector_attribute_name: The property name for vector indexed property on the searched object.
    :type vector_attribute_name: str
    :param candidate_vector: The vector you are finding the nearest topk neighbours for. 
    :type candidate_vector: list[float]

    """
    def __init__(self, topk: int, vector_attribute_name: str, candidate_vector: List[float]):
        self.topk = topk
        self.vector_attribute_name = vector_attribute_name
        self.index_name = None 
        self.nodeSetLabel = None
        self.vector = candidate_vector
        
